#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2022 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

from pathlib import Path
import subprocess
import re
import time, sys, os
import argparse
import shutil
import dmgbuild

thisFolder = Path(__file__).parent
finalDistFolder = thisFolder.parent.parent/'dist'

with Path().home()/ 'keys/apple_ost_id' as p:
    IDENTITY = p.read_text().strip()
with Path().home()/ 'keys/apple_psychopy_app_specific' as p:
    PWORD = p.read_text().strip()

ENTITLEMENTS = thisFolder / "entitlements.plist"
BUNDLE_ID = "org.opensciencetools.psychopy"
USERNAME = "admin@opensciencetools.org"

SIGN_ALL = True

# handy resources for info:
#
# why use zip file to notarize as well as dmg:
#   https://deciphertools.com/blog/notarizing-dmg/
# notarize from Python:
#   https://github.com/najiji/notarizer/blob/master/notarize.py
# apple entitlements:
#     https://developer.apple.com/documentation/xcode/notarizing_macos_software_before_distribution/resolving_common_notarization_issues


class AppSigner:
    def __init__(self, appFile, version, destination=None):
        self.appFile = Path(appFile)
        self.version = version
        self.destination = destination
        self._zipFile = None #'/Users/lpzjwp/code/psychopy/git/dist/PsychoPy3_2020.2.3.zip'
        self._appNotarizeUUID = None
        self._dmgBuildFile = None

    def signAll(self):
        # remove files that we know will fail the signing:
        for filename in self.appFile.glob("**/Frameworks/SDL*"):
            shutil.rmtree(filename)
        for filename in self.appFile.glob("**/Frameworks/eyelink*"):
            shutil.rmtree(filename)

        # this never really worked - probably the files signed in wrong order?
        # find all the included dylibs
        print('Signing dylibs:', end='')
        files = list(self.appFile.glob('**/*.dylib'))
        files.extend(self.appFile.glob('**/*.so'))
        files.extend(self.appFile.glob('**/git-core/git*'))
        files.extend(self.appFile.glob('**/cv2/.dylibs/*'))
        # ffmpeg
        files.extend(self.appFile.glob('**/imageio_ffmpeg/binaries/*'))
        files.extend(self.appFile.glob('**/resources/ffmpeg/ffmpeg-osx*'))
        # PyQt
        files.extend(self.appFile.glob('**/Versions/5/Qt*'))
        files.extend(self.appFile.glob('**/Contents/MacOS/QtWebEngineProcess'))
        files.extend(self.appFile.glob('**/Resources/lib/python3.6/lib-dynload/*.so'))
        files.extend(self.appFile.glob('**/Frameworks/Python.framework/Versions/3.6/Python'))
        files.extend(self.appFile.glob('**/Frameworks/Python.framework'))
        files.extend(self.appFile.glob('**/Contents/MacOS/python'))

        # ready? Let's do this!
        t0 = time.time()
        for filename in files:
            print('.', end='')
            sys.stdout.flush()
            if filename.exists():  # might have been removed since glob
                self.signSingleFile(filename, verbose=False, removeFailed=True)
        print(f'...done signing dylibs in {time.time()-t0:.03f}s')

        # then sign the outer app file
        print('Signing app')
        sys.stdout.flush()
        t0 = time.time()
        self.signSingleFile(self.appFile, removeFailed=False)
        print(f'...done signing app in {time.time()-t0:.03f}s')
        sys.stdout.flush()

    def signSingleFile(self, filename, removeFailed=False, verbose=True,
                       appFile=False):
        cmd = ['codesign', str(filename),
               '--sign',  IDENTITY,
               '--entitlements', str(ENTITLEMENTS),
               '--force',
               '--timestamp',
               # '--deep',  # not recommended although used in most demos
               '--options', 'runtime',
               ]
        cmdStr = ' '.join(cmd)
        if verbose:
            print(cmdStr)
        exitcode, output = subprocess.getstatusoutput(cmdStr)
        # if failed or verbose then give info
        if exitcode != 0 or ('failed' in output) or (verbose and output):
            print(output)
        # if failed and removing then remove
        if (exitcode != 0 or 'failed' in output) and removeFailed:
            Path(filename).unlink()
            print(f"REMOVED FILE {filename}: failed to codesign")
        return self.signCheck(filename, verbose=False, removeFailed=removeFailed)

    def signCheck(self, filepath=None, verbose=False, strict=True,
                  removeFailed=False):
        """Checks whether a file is signed and returns a list of warnings"""
        if not filepath:
            filepath = self.appFile
        # just check the details
        strictFlag = "--strict" if strict else ""
        cmdStr = f'codesign -dvvv {strictFlag} {filepath}'
        # make the call
        if verbose:
            print(cmdStr)
        exitcode, output = subprocess.getstatusoutput(cmdStr)
        if verbose:
            print(f"Checking that codesign worked: {output}")

        # check for warnings
        warnings=[]
        for line in output.split("\n"):
            if 'warning' in line.lower():
                warnings.append(line)
        if warnings:
            print(filepath)
            for line in warnings:
                print("  ", line)
            if removeFailed:
                Path(filepath).unlink()
                print(f"REMOVED FILE {filepath}: failed to codesign")
        return warnings

    def upload(self, fileToNotarize):
        filename = Path(fileToNotarize).name
        print(f'Sending {filename} to apple for notarizing')
        cmdStr = (f"xcrun altool --notarize-app -t osx -f {fileToNotarize} "
                  f"--primary-bundle-id {BUNDLE_ID} -u {USERNAME} ")
        print(cmdStr)
        cmdStr += f"-p {PWORD}"
        t0 = time.time()
        exitcode, output = subprocess.getstatusoutput(cmdStr)
        m = re.match('.*RequestUUID = (.*)\n', output, re.S)
        if 'Please sign in with an app-specific password' in output:
            print("[Error] Upload failed: You probably need a new app-specific "
                  "password from https://appleid.apple.com/account/manage")
            exit(1)
        elif m is None or not ('No errors uploading' in output):
            print(f'[Error] Upload failed: {output}')
            exit(1)
        uuid = m.group(1).strip()
        self._appNotarizeUUID = uuid
        print(f'Uploaded file {filename} in {time.time()-t0:.03f}s: {uuid}')
        print(f'Upload to Apple completed at {time.ctime()}')
        sys.stdout.flush()
        return uuid

    @property
    def dmgFile(self):
        if not self._dmgBuildFile:
            self._dmgBuildFile = self._buildDMG()
        return self._dmgBuildFile

    @property
    def zipFile(self):
        if self._zipFile:
            return self._zipFile
        else:
            print("Creating zip archive to send to Apple: ", end='')
            zipFilename = self.appFile.parent / (self.appFile.stem+f'_{self.version}.zip')
            shutil.rmtree(zipFilename, ignore_errors=True)
            # zipFilename.unlink(missing_ok=True)  # remove the file if it exists
            t0 = time.time()
            cmdStr = f'/usr/bin/ditto -c -k --keepParent {self.appFile} {zipFilename}'
            print(cmdStr)
            exitcode, output = subprocess.getstatusoutput(cmdStr)
            if exitcode == 0:
                print(f"Done creating zip in {time.time()-t0:.03f}s")
            else:
                print(output)
            self._zipFile = zipFilename
            return zipFilename

    def awaitNotarized(self):
        while self.checkStatus(self._appNotarizeUUID):  # returns True while in progress
            time.sleep(30)


    def checkStatus(self, uuid):
        cmd = ['xcrun', 'altool', '--notarization-info', self._appNotarizeUUID,
               '-u', USERNAME, '-p', PWORD]
        cmdStr = ' '.join(cmd)
        exitcode, output = subprocess.getstatusoutput(cmdStr)

        in_progress = 'Status: in progress' in output
        success = 'Status: success' in output

        if not in_progress:
            print(f'Notarization completed at {time.ctime()}')
            if not success:
                print('*********Notarization failed*************')
                print(output)
                exit(1)

        return in_progress

    def staple(self, filepath):
        cmdStr = f'xcrun stapler staple {filepath}'
        print(cmdStr)
        exitcode, output = subprocess.getstatusoutput(cmdStr)
        print(f"exitcode={exitcode}: {output}")
        if exitcode != 0:
            print('*********Staple failed*************')
            exit(1)
        else:
            print(f"Staple successful. You can verify with\n    xcrun stapler validate {filepath}")

    def checkAppleLogFile(self):
        cmdStr = f"xcrun altool --notarization-info {self._appNotarizeUUID} -u {USERNAME} -p {PWORD}"

        exitcode, output = subprocess.getstatusoutput(cmdStr)
        print(f"exitcode={exitcode}: {output}")

    def dmgBuild(self):
        dmgFilename = str(self.appFile).replace(".app", "_rw.dmg")
        appName = self.appFile.name
        print(f"building dmg file: {dmgFilename}")
        # remove previous file if it's there
        if Path(dmgFilename).exists():
            os.remove(dmgFilename)
        # then build new one

        icon = (thisFolder.parent /
                'psychopy/app/Resources/psychopy.icns').resolve()
        background = (thisFolder / "dmg722x241.tiff").resolve()
        dmgbuild.build_dmg(
                filename=dmgFilename,
                volume_name=f'PsychoPy-{self.version}',  # avoid spaces
                settings={
                    'format': 'UDRW',
                    'files': [str(self.appFile)],
                    'symlinks': { 'Applications': '/Applications' },
                    'size': '3g',  # but maybe irrelevant in UDRW mode?
                    'badge_icon': str(icon),
                    'background': None,  # background
                    'icon_size': 128,
                    'icon_locations': {
                        'PsychoPy.app': (150, 160),
                        'Applications': (350, 160)
                    },
                    'window_rect': ((600, 600), (500, 400)),
                },
        )
        print(f"building dmg file complete")
        sys.stdout.flush()
        return dmgFilename

    def dmgStapleInside(self):
        dmgFilename = str(self.appFile).replace(".app", "_rw.dmg")
        appName = self.appFile.name
        """Staple the notarization to the app inside the r/w dmg file"""
        # staple the file inside the dmg
        cmdStr = f"hdiutil attach '{dmgFilename}'"
        exitcode, output = subprocess.getstatusoutput(cmdStr)
        # subprocess.getstatusoutput("say 'waiting' --voice=Kate")
        time.sleep(10)
        volName = output.split('\t')[-1]
        self.staple(f"'{volName}/{appName}'")
        cmdStr = f"hdiutil detach '{volName}' -quiet"
        print(f'cmdStr was: {cmdStr}')
        for n in range(5):  # if we do this too fast then it fails. Try 5 times
            time.sleep(10)
            exitcode, output = subprocess.getstatusoutput(cmdStr)
            print(output)
            if exitcode == 0:
                break  # succeeded so stop
        if exitcode != 0:
            print(f'*********Failed to detach {volName} (wrong name?) *************')
            exit(1)

    def dmgCompress(self):
        dmgFilename = str(self.appFile).replace(".app", "_rw.dmg")
        dmgFinalFilename = finalDistFolder/(f"StandalonePsychoPy-{self.version}-macOS.dmg")
        # remove previous file if it's there
        if Path(dmgFinalFilename).exists():
            os.remove(dmgFinalFilename)

        cmdStr = f"hdiutil convert {dmgFilename} " \
                 f"-format UDZO " \
                 f"-o {dmgFinalFilename}"
        exitcode, output = subprocess.getstatusoutput(cmdStr)
        print(output)
        if exitcode != 0:
            print(f'****Failed to compress {dmgFilename} to {dmgFinalFilename} (is it not ejected?) ****')
            exit(1)
        return dmgFinalFilename


def main():

    with open(thisFolder.parent / "version") as f:
        defaultVersion = f.read().strip()
    parser = argparse.ArgumentParser(description="Codesigning PsychoPy.app")
    parser.add_argument("--app", help=("Path to the app bundle, "
                                       "assumed to be in dist/"),
                        action='store', required=False, default="PsychoPy.app")
    parser.add_argument("--version", help="Version of the app",
                        action='store', required=False, default=defaultVersion)
    parser.add_argument("--file", help="path for a single file to be signed",
                        action='store', required=False, default=None)
    parser.add_argument("--skipnotarize", help="path for a single file to be signed",
                        action='store', required=False, default=None)
    parser.add_argument("--runPreDmgBuild", help="Runs up until dmg is built (and notarised) then exits",
                        action='store', required=False, default='true')
    parser.add_argument("--runDmgBuild", help="Runs the dmg build itself",
                        action='store', required=False, default='true')
    parser.add_argument("--runPostDmgBuild", help="Runs up until dmg is built (and notarised) then exits",
                        action='store', required=False, default='true')
    args = parser.parse_args()
    args.runPreDmgBuild = args.runPreDmgBuild.lower() in ['true', 'True', '1', 'y', 'yes']
    args.runDmgBuild = args.runDmgBuild.lower() in ['true', 'True', '1', 'y', 'yes']
    args.runPostDmgBuild = args.runPostDmgBuild.lower() in ['true', 'True', '1', 'y', 'yes']

    if args.skipnotarize:
        NOTARIZE = False
    else:
        NOTARIZE = True

    if args.file:  # not the whole app - just sign one file
        distFolder = (thisFolder / '../dist').resolve()
        signer = AppSigner(appFile='',
                           version=None)
        signer.signSingleFile(args.file, removeFailed=False, verbose=True)
        signer.signCheck(args.file, verbose=True)

        if NOTARIZE:
            signer.upload(args.file)
            # notarize and staple
            signer.awaitNotarized()
            signer.staple(args.file)

    else:  # full app signing and notarization
        distFolder = (thisFolder / '../dist').resolve()
        signer = AppSigner(appFile=distFolder/args.app,
                        version=args.version)

        if args.runPreDmgBuild:
            if SIGN_ALL:
                signer.signAll()
            signer.signCheck(verbose=False)

            if NOTARIZE and args.runDmgBuild:
                signer.upload(signer.zipFile)
                # build the read/writable dmg file while waiting for notarize
                signer.dmgBuild()
                # notarize and staple
                signer.awaitNotarized()
            elif args.runDmgBuild:
                # just build the dmg
                signer.dmgBuild()

        if args.runPostDmgBuild:
            signer.dmgStapleInside()  # doesn't require UUID

            dmgFile = signer.dmgCompress()
            signer.signSingleFile(dmgFile, removeFailed=False, verbose=True)

            if NOTARIZE:
                signer.upload(dmgFile)
                # notarize and staple
                signer.awaitNotarized()
                signer.staple(dmgFile)


if __name__ == "__main__":
    main()



import os
import shutil
from pathlib import Path
from tempfile import mkdtemp
from psychopy.experiment import getAllComponents, Experiment
from psychopy.tests.utils import compareTextFiles, TESTS_DATA_PATH
from psychopy.scripts import psyexpCompile


class TestComponentCompilerPython():
    """A class for testing the Python code compiler for all components"""
    def setup(self):
        self.temp_dir = mkdtemp()
        self.allComp = getAllComponents(fetchIcons=False)
        self.exp = Experiment() # create once, not every test
        self.exp.addRoutine('trial')
        self.exp.flow.addRoutine(self.exp.routines['trial'], pos=0)
        # Create correctScript subdir for holding correct scripts
        if not os.path.isdir(os.path.join(TESTS_DATA_PATH, "correctScript", "python")):
            os.mkdir(os.path.join(TESTS_DATA_PATH, "correctScript", "python"))

    def teardown(self):
        shutil.rmtree(self.temp_dir)

    def test_all_components(self):
        """Test all component code outputs, except for Settings and Unknown"""
        for compName in self.allComp:
            if compName not in ['SettingsComponent', 'UnknownComponent']:
                # reset exp
                self.reset_experiment()
                # Add components
                self.add_components(compName)
                # Create output script
                self.create_component_output(compName)
                # Get correct script path
                # correctPath = os.path.join(TESTS_DATA_PATH, "correctScript", "python", 'correct{}.py'.format(compName))
                # Compare files, raising assertions on fails above tolerance (%)
                # try:
                #     compareTextFiles('new{}.py'.format(compName), correctPath, tolerance=5)
                # except IOError as err:
                #     compareTextFiles('new{}.py'.format(compName), correctPath, tolerance=5)

    def reset_experiment(self):
        """Resets the exp object for each component"""
        self.exp = Experiment()
        self.exp.addRoutine('trial')
        self.exp.flow.addRoutine(self.exp.routines['trial'], pos=0)

    def add_components(self, compName):
        """Add components to routine"""
        thisComp = self.allComp[compName](parentName='trial', exp=self.exp)
        if compName == 'StaticComponent':
            # Create another component to trigger param updates for static
            textStim = self.allComp['TextComponent'](parentName='trial', exp=self.exp)
            textStim.params['color'].allowedUpdates.append('set during: trial.ISI')
            textStim.params['color'].updates = 'set during: trial.ISI'
            self.exp.routines['trial'].addComponent(textStim)
            # Create static component
            thisComp.addComponentUpdate('trial', 'text', 'color')
            thisComp.params['code'].val = "customStaticCode = True"  # Add the custom code
            self.exp.routines['trial'].addComponent(thisComp)
        else:
            self.exp.routines['trial'].addComponent(thisComp)

    def create_component_output(self, compName):
        """Create the Python script"""
        pyFilePath = os.path.join(self.temp_dir, 'new{}.py'.format(compName))
        psyexpCompile.compileScript(infile=self.exp, outfile=pyFilePath)

    def test_component_type_in_experiment(self):
        for compName in self.allComp:
            if compName not in ['SettingsComponent', 'UnknownComponent']:
                # reset exp
                self.reset_experiment()
                # Add components
                self.add_components(compName)
                # Check component in exp
                component = compName.split('Component')[0]
                assert self.exp.getComponentFromType(component), (
                    f"Could not find component of type {compName} in: {self.exp.flow}"
                )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
ElementArray demo.

This demo requires a graphics card that supports OpenGL2 extensions.

It shows how to manipulate an arbitrary set of elements using numpy arrays
and avoiding for loops in your code for optimised performance.

See also the elementArrayStim demo.
"""

from psychopy import visual, event, core
from psychopy.tools.coordinatetools import pol2cart
import numpy

nDots = 500
maxSpeed = 0.02
dotSize = .0075

dotsTheta = numpy.random.rand(nDots) * 360
dotsRadius = (numpy.random.rand(nDots) ** 0.5) * 2
speed = numpy.random.rand(nDots) * maxSpeed

win = visual.Window([800, 600], color=[-1, -1, -1])
dots = visual.ElementArrayStim(win, elementTex=None, elementMask='circle',
    nElements=nDots, sizes=dotSize)

while not event.getKeys():
    # update radius
    dotsRadius = (dotsRadius + speed)
    # random radius where radius too large
    outFieldDots = (dotsRadius >= 2.0)
    dotsRadius[outFieldDots] = numpy.random.rand(sum(outFieldDots)) * 2.0

    dotsX, dotsY = pol2cart(dotsTheta, dotsRadius)
    dotsX *=  0.75  # to account for wider aspect ratio
    dots.xys = numpy.array([dotsX, dotsY]).transpose()
    dots.draw()

    win.flip()

win.close()
core.quit()

# The contents of this file are in the public domain.



#!/usr/bin/env python
# coding: utf-8

"""
Split string in accordance with UAX#14 Unicode line breaking.

Code is based on uniseg 0.7.1 (https://pypi.org/project/uniseg/)
"""

import sys
import re
from psychopy.tools.linebreak_class import linebreak_class

__all__ = [
    'get_breakable_points',
    'break_units',
]

from builtins import ord as _ord


if sys.maxunicode < 0x10000:
    # narrow unicode build
    def ord(c, index=None):
        if isinstance(c, str):
            return _ord(c[index or 0])
        # if not isinstance(c, unicode):
        #     raise TypeError('must be unicode, not %s' % type(c).__name__)
        i = index or 0
        len_s = len(c)-i
        if len_s:
            value = hi = _ord(c[i])
            i += 1
            if 0xd800 <= hi < 0xdc00:
                if len_s > 1:
                    lo = _ord(c[i])
                    i += 1
                    if 0xdc00 <= lo < 0xe000:
                        value = (hi-0xd800)*0x400+(lo-0xdc00)+0x10000
            if index is not None or i == len_s:
                return value
        raise TypeError('need a single Unicode code point as parameter')

    rx_codepoints = re.compile(r'[\ud800-\udbff][\udc00-\udfff]|.', re.DOTALL)
    
    def code_point(s, index=0):
        
        L = rx_codepoints.findall(s)
        return L[index]
    
    def code_points(s):
        
        return rx_codepoints.findall(s)

else:
    # wide unicode build
    
    def ord(c, index=None):
        return _ord(c if index is None else c[index])
    
    def code_point(s, index=0):
        return s[index or 0]
    
    def code_points(s):
        return list(s)

BK = 'BK'   # Mandatory Break
CR = 'CR'   # Carriage Return
LF = 'LF'   # Line Feed
CM = 'CM'   # Combining Mark
NL = 'NL'   # Next Line
SG = 'SG'   # Surrogate
WJ = 'WJ'   # Word Joiner
ZW = 'ZW'   # Zero Width Space
GL = 'GL'   # Non-breaking ("Glue")
SP = 'SP'   # Space
B2 = 'B2'   # Break Opportunity Before and After
BA = 'BA'   # Break After
BB = 'BB'   # Break Before
HY = 'HY'   # Hyphen
CB = 'CB'   # Contingent Break Opportunity
CL = 'CL'   # Close Punctuation
CP = 'CP'   # Close Parenthesis
EX = 'EX'   # Exclamation/Interrogation
IN = 'IN'   # Inseparable
NS = 'NS'   # Nonstarter
OP = 'OP'   # Open Punctuation
QU = 'QU'   # Quotation
IS = 'IS'   # Infix Numeric Separator
NU = 'NU'   # Numeric
PO = 'PO'   # Postfix Numeric
PR = 'PR'   # Prefix Numeric
SY = 'SY'   # Symbols Allowing Break After
AI = 'AI'   # Ambiguous (Alphabetic or Ideographic)
AL = 'AL'   # Alphabetic
CJ = 'CJ'   # Conditional Japanese Starter
H2 = 'H2'   # Hangul LV Syllable
H3 = 'H3'   # Hangul LVT Syllable
HL = 'HL'   # Hebrew Letter
ID = 'ID'   # Ideographic
JL = 'JL'   # Hangul L Jamo
JV = 'JV'   # Hangul V Jamo
JT = 'JT'   # Hangul T Jamo
RI = 'RI'   # Regional Indicator
SA = 'SA'   # Complex Context Dependent (South East Asian)
XX = 'XX'   # Unknown

def line_break(c, index=0):
    code = ord(code_point(c, index))
    if code in linebreak_class:
        return linebreak_class[code]
    return 'Other'

def break_units(s, breakables):
    """
    Split a sequence at given breakpoint. This returns a generator object.
    So do `list(break_units(s, breakables))` to get the result as a list.
    
    :Parameters:
    
        s:
            A string (or sequence) to be split.

        breakables:
            A sequence of 0/1 of the same length of s. 1 represents that 
            the input sequence is breakable at that point.
            See also get_breakable_points().
    """
    i = 0
    for j, bk in enumerate(breakables):
        if bk:
            if j:
                yield s[i:j]
            i = j
    if s:
        yield s[i:]

def _preprocess_boundaries(s):
    prev_prop = None
    i = 0
    for c in code_points(s):
        prop = line_break(c)
        if prop in (BK, CR, LF, SP, NL, ZW):
            yield (i, prop)
            prev_prop = None
        elif prop == CM:
            if prev_prop is None:
                yield (i, prop)
                prev_prop = prop
        else:
            yield (i, prop)
            prev_prop = prop
        i += len(c)

def get_breakable_points(s):
    """
    Returns a generator object that yields 1 if the next character is
    breakable, otherwise yields 0.
    Do `list(get_breakable_points(s))` to get a list of breakable points.
    
    :Parameters:
    
        s:
            Sentence to be parsed.
    """
    if not s:
        return
    
    primitive_boundaries = list(_preprocess_boundaries(s))
    prev_prev_lb = None
    prev_lb = None
    for i, (pos, lb) in enumerate(primitive_boundaries):
        next_pos, __ = (primitive_boundaries[i+1]
                        if i<len(primitive_boundaries)-1 else (len(s), None))
        
        if lb == AI:
            lb = AL
        
        if lb == CJ:
            lb = NS

        if lb in (CM, XX, SA):
            lb = AL
        # LB4
        if pos == 0:
            do_break = False
        elif prev_lb == BK:
            do_break = True
        # LB5
        elif prev_lb in (CR, LF, NL):
            do_break = not (prev_lb == CR and lb == LF)
        # LB6
        elif lb in (BK, CR, LF, NL):
            do_break = False
        # LB7
        elif lb in (SP, ZW):
            do_break = False
        # LB8
        elif ((prev_prev_lb == ZW and prev_lb == SP) or (prev_lb == ZW)):
            do_break = True
        # LB11
        elif lb == WJ or prev_lb == WJ:
            do_break = False
        # LB12
        elif prev_lb == GL:
            do_break = False
        # LB12a
        elif prev_lb not in (SP, BA, HY) and lb == GL:
            do_break = False
        # LB13
        elif lb in (CL, CP, EX, IS, SY):
            do_break = False
        # LB14
        elif (prev_prev_lb == OP and prev_lb == SP) or prev_lb == OP:
            do_break = False
        # LB15
        elif ((prev_prev_lb == QU and prev_lb == SP and lb == OP)
              or (prev_lb == QU and lb == OP)):
            do_break = False
        # LB16
        elif ((prev_prev_lb in (CL, CP) and prev_lb == SP and lb == NS)
              or (prev_lb in (CL, CP) and lb == NS)):
            do_break = False
        # LB17
        elif ((prev_prev_lb == B2 and prev_lb == SP and lb == B2)
              or (prev_lb == B2 and lb == B2)):
            do_break = False
        # LB18
        elif prev_lb == SP:
            do_break = True
        # LB19
        elif lb == QU or prev_lb == QU:
            do_break = False
        # LB20
        elif lb == CB or prev_lb == CB:
            do_break = True
        # LB21
        elif lb in (BA, HY, NS) or prev_lb == BB:
            do_break = False
        # LB22
        elif prev_lb in (AL, HL, ID, IN, NU) and lb == IN:
            do_break = False
        # LB23
        elif ((prev_lb == ID and lb == PO)
              or (prev_lb in (AL, HL) and lb == NU)
              or (prev_lb == NU and lb in (AL, HL))):
            do_break = False
        # LB24
        elif ((prev_lb == PR and lb == ID)
              or (prev_lb == PR and lb in (AL, HL))
              or (prev_lb == PO and lb in (AL, HL))):
            do_break = False
        # LB25
        elif ((prev_lb == CL and lb == PO)
              or (prev_lb == CP and lb == PO)
              or (prev_lb == CL and lb == PR)
              or (prev_lb == CP and lb == PR)
              or (prev_lb == NU and lb == PO)
              or (prev_lb == NU and lb == PR)
              or (prev_lb == PO and lb == OP)
              or (prev_lb == PO and lb == NU)
              or (prev_lb == PR and lb == OP)
              or (prev_lb == PR and lb == NU)
              or (prev_lb == HY and lb == NU)
              or (prev_lb == IS and lb == NU)
              or (prev_lb == NU and lb == NU)
              or (prev_lb == SY and lb == NU)):
            do_break = False
        # LB26
        elif ((prev_lb == JL and lb in (JL, JV, H2, H3))
              or (prev_lb in (JV, H2) and lb in (JV, JT))
              or (prev_lb in (JT, H3) and lb == JT)):
            do_break = False
        # LB27
        elif ((prev_lb in (JL, JV, JT, H2, H3) and lb in (IN, PO))
              or (prev_lb == PR and lb in (JL, JV, JT, H2, H3))):
            do_break = False
        # LB28
        elif prev_lb in (AL, HL) and lb in (AL, HL):
            do_break = False
        # LB29
        elif prev_lb == IS and lb in (AL, HL):
            do_break = False
        # LB30
        elif ((prev_lb in (AL, HL, NU) and lb == OP)
              or (prev_lb == CP and lb in (AL, HL, NU))):
            do_break = False
        # LB30a
        elif prev_lb == lb == RI:
            do_break = False
        else:
            do_break = True
        for j in range(next_pos-pos):
            yield int(j==0 and do_break)
        prev_prev_lb = prev_lb
        prev_lb = lb



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2022 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

"""Experiment classes:
    Experiment, Flow, Routine, Param, Loop*, *Handlers, and NameSpace

The code that writes out a *_lastrun.py experiment file is (in order):
    experiment.Experiment.writeScript() - starts things off, calls other parts
    settings.SettingsComponent.writeStartCode()
    experiment.Flow.writeBody()
        which will call the .writeBody() methods from each component
    settings.SettingsComponent.writeEndCode()
"""

from .params import getCodeFromParamStr, Param
from .components import getInitVals, getComponents, getAllComponents
from .routines import getAllStandaloneRoutines
from ._experiment import Experiment
from .utils import unescapedDollarSign_re, valid_var_re, \
     nonalphanumeric_re
from psychopy.experiment.utils import CodeGenerationException



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2022 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

from pathlib import Path
from psychopy.alerts import alerttools
from psychopy.experiment.components import BaseVisualComponent, Param, getInitVals, _translate
from psychopy.localization import _localized as __localized
_localized = __localized.copy()

# only use _localized values for label values, nothing functional:
_localized.update({'text': _translate('Text'),
                   'font': _translate('Font'),
                   'letterHeight': _translate('Letter height'),
                   'wrapWidth': _translate('Wrap width'),
                   'flip': _translate('Flip (mirror)'),
                   'languageStyle': _translate('Language style')})


class TextComponent(BaseVisualComponent):
    """An event class for presenting text-based stimuli
    """

    categories = ['Stimuli']
    targets = ['PsychoPy', 'PsychoJS']
    iconFile = Path(__file__).parent / 'text.png'
    tooltip = _translate('Text: present text stimuli')

    def __init__(self, exp, parentName, name='text',
                 # effectively just a display-value
                 text=_translate('Any text\n\nincluding line breaks'),
                 font='Open Sans', units='from exp settings',
                 color='white', colorSpace='rgb',
                 pos=(0, 0), letterHeight=0.05, ori=0,
                 startType='time (s)', startVal=0.0,
                 stopType='duration (s)', stopVal=1.0,
                 flip='None', startEstim='', durationEstim='', wrapWidth='',
                 languageStyle='LTR'):
        super(TextComponent, self).__init__(exp, parentName, name=name,
                                            units=units,
                                            color=color,
                                            colorSpace=colorSpace,
                                            pos=pos,
                                            ori=ori,
                                            startType=startType,
                                            startVal=startVal,
                                            stopType=stopType,
                                            stopVal=stopVal,
                                            startEstim=startEstim,
                                            durationEstim=durationEstim)
        self.type = 'Text'
        self.url = "https://www.psychopy.org/builder/components/text.html"

        # params
        _allow3 = ['constant', 'set every repeat', 'set every frame']  # list
        self.params['text'] = Param(
            text, valType='str', inputType="multi", allowedTypes=[], categ='Basic',
            updates='constant', allowedUpdates=_allow3[:],  # copy the list
            hint=_translate("The text to be displayed"),
            canBePath=False,
            label=_localized['text'])
        self.params['font'] = Param(
            font, valType='str', inputType="single", allowedTypes=[], categ='Formatting',
            updates='constant', allowedUpdates=_allow3[:],  # copy the list
            hint=_translate("The font name (e.g. Comic Sans)"),
            label=_localized['font'])
        del self.params['size']  # because you can't specify width for text
        self.params['letterHeight'] = Param(
            letterHeight, valType='num', inputType="single", allowedTypes=[], categ='Formatting',
            updates='constant', allowedUpdates=_allow3[:],  # copy the list
            hint=_translate("Specifies the height of the letter (the width"
                            " is then determined by the font)"),
            label=_localized['letterHeight'])

        self.params['wrapWidth'] = Param(
            wrapWidth, valType='num', inputType="single", allowedTypes=[], categ='Layout',
            updates='constant', allowedUpdates=['constant'],
            hint=_translate("How wide should the text get when it wraps? (in"
                            " the specified units)"),
            label=_localized['wrapWidth'])
        self.params['flip'] = Param(
            flip, valType='str', inputType="single", allowedTypes=[], categ='Layout',
            allowedVals=["horiz", "vert", "None"], updates='constant', allowedUpdates=_allow3[:],  # copy the list
            hint=_translate("horiz = left-right reversed; vert = up-down"
                            " reversed; $var = variable"),
            label=_localized['flip'])
        self.params['languageStyle'] = Param(
            languageStyle, valType='str', inputType="choice", categ='Formatting',
            allowedVals=['LTR', 'RTL', 'Arabic'],
            hint=_translate("Handle right-to-left (RTL) languages and Arabic reshaping"),
            label=_localized['languageStyle'])

        del self.params['fillColor']
        del self.params['borderColor']

    def writeInitCode(self, buff):
        # do we need units code?
        if self.params['units'].val == 'from exp settings':
            unitsStr = ""
        else:
            unitsStr = "units=%(units)s, " % self.params
        # do writing of init
        # replaces variable params with sensible defaults
        inits = getInitVals(self.params, 'PsychoPy')
        if self.params['wrapWidth'].val in ['', 'None', 'none']:
            inits['wrapWidth'] = 'None'

        code = ("%(name)s = visual.TextStim(win=win, "
                "name='%(name)s',\n"
                "    text=%(text)s,\n"
                "    font=%(font)s,\n"
                "    " + unitsStr +
                "pos=%(pos)s, height=%(letterHeight)s, "
                "wrapWidth=%(wrapWidth)s, ori=%(ori)s, \n"
                "    color=%(color)s, colorSpace=%(colorSpace)s, "
                "opacity=%(opacity)s, \n"
                "    languageStyle=%(languageStyle)s,")
        buff.writeIndentedLines(code % inits)
        flip = self.params['flip'].val.strip()
        if flip == 'horiz':
            flipStr = 'flipHoriz=True, '
        elif flip == 'vert':
            flipStr = 'flipVert=True, '
        else:
            flipStr = ''
        depth = -self.getPosInRoutine()
        buff.writeIndented('    ' + flipStr + 'depth=%.1f);\n' % depth)

    def writeInitCodeJS(self, buff):
        # do we need units code?
        if self.params['units'].val == 'from exp settings':
            unitsStr = "  units: undefined, \n"
        else:
            unitsStr = "  units: %(units)s, \n" % self.params
        # do writing of init
        # replaces variable params with sensible defaults
        inits = getInitVals(self.params, 'PsychoJS')

        # check for NoneTypes
        for param in inits:
            if inits[param] in [None, 'None', '']:
                inits[param].val = 'undefined'
                if param == 'text':
                    inits[param].val = ""

        code = ("%(name)s = new visual.TextStim({\n"
                "  win: psychoJS.window,\n"
                "  name: '%(name)s',\n"
                "  text: %(text)s,\n"
                "  font: %(font)s,\n" + unitsStr +
                "  pos: %(pos)s, height: %(letterHeight)s,"
                "  wrapWidth: %(wrapWidth)s, ori: %(ori)s,\n"
                "  color: new util.Color(%(color)s),"
                "  opacity: %(opacity)s,")
        buff.writeIndentedLines(code % inits)

        flip = self.params['flip'].val.strip()
        if flip == 'horiz':
            flipStr = 'flipHoriz : true, '
        elif flip == 'vert':
            flipStr = 'flipVert : true, '
        elif flip and not flip == "None":
            msg = ("flip value should be 'horiz' or 'vert' (no quotes)"
                   " in component '%s'")
            raise ValueError(msg % self.params['name'].val)
        else:
            flipStr = ''
        depth = -self.getPosInRoutine()
        code = ("  %sdepth: %.1f \n"
                "});\n\n" % (flipStr, depth))
        buff.writeIndentedLines(code)

    def integrityCheck(self):
        super().integrityCheck()  # run parent class checks first
        alerttools.testFont(self)  # Test whether font is available locally



# -*- coding: utf-8 -*-

from psychopy import info, visual
import pytest

# py.test -k info --cov-report term-missing --cov info.py


@pytest.mark.info
class TestInfo():
    @classmethod
    def setup_class(self):
        self.win = visual.Window(size=(100,100), autoLog=False)
    def teardown(self):
        self.win.close()

    def test_info(self):
        info.RunTimeInfo(win=self.win, userProcsDetailed=True, verbose=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Demo of using the iohub mouse.
"""

import sys

from psychopy import visual, core
from psychopy.iohub import launchHubServer


win = visual.Window((1920, 1080), units='height', fullscr=True, allowGUI=False, screen=0)

# create the process that will run in the background polling devices
io = launchHubServer(window=win)

# some default devices have been created that can now be used
keyboard = io.devices.keyboard
mouse = io.devices.mouse
mouse.setPosition((0.0, .250))
#win.setMouseVisible(False)

# Create some psychopy visual stim
fixSpot = visual.GratingStim(win, tex="none", mask="gauss", pos=(0, 0), size=(.03, .03), color='black', autoLog=False)
grating = visual.GratingStim(win, pos=(.3, 0), tex="sin", mask="gauss", color=[1.0, .5, -1.0], size=(.15, .15),
                             sf=(.01, 0), autoLog=False)
message = visual.TextStim(win, pos=(0, -.2), height=.03, alignText='center', anchorHoriz='center', anchorVert='center',
                          wrapWidth=.7, text='move=mv-spot, left-drag=SF, right-drag=mv-grating, scroll=ori',
                          autoLog=False)
displayIdMsg = visual.TextStim(win, pos=(0.0, -0.3), alignText='center', anchorHoriz='center', anchorVert='center',
                               height=.03, text='Display X', autoLog=False, wrapWidth=0.7)
message3 = visual.TextStim(win, pos=(0.0, -0.4), alignText='center', anchorHoriz='center', anchorVert='center',
                           height=.03, text='Press Any Key to Quit.', autoLog=False, wrapWidth=0.7)
last_wheelPosY = 0

io.clearEvents('all')

demo_timeout_start = core.getTime()
# Run the example until a keyboard event is received.

kb_events = None
last_display_ix = -1
while not kb_events:
    # Get the current mouse position
    # posDelta is the change in position * since the last call *
    position, posDelta, display_ix = mouse.getPositionAndDelta(return_display_index=True)
    mouse_dX, mouse_dY = posDelta

    if display_ix is not None and display_ix != last_display_ix:
        displayIdMsg.setText("Display %d" % display_ix)
        last_display_ix = display_ix

    # Get the current state of each of the Mouse Buttons
    left_button, middle_button, right_button = mouse.getCurrentButtonStates()

    # If the left button is pressed, change the grating's spatial frequency
    if left_button:
        grating.setSF(mouse_dX / 5000.0, '+')
    elif right_button:
        grating.setPos(position)

    # If no buttons are pressed on the Mouse, move the position of the mouse cursor.
    if True not in (left_button, middle_button, right_button):
        fixSpot.setPos(position)

    if sys.platform == 'darwin':
        # On macOS, both x and y mouse wheel events can be detected, assuming the mouse being used
        # supported 2D mouse wheel motion.
        wheelPosX, wheelPosY = mouse.getScroll()
    else:
        # On Windows and Linux, only vertical (Y) wheel position is supported.
        wheelPosY = mouse.getScroll()

    # keep track of the wheel position 'delta' since the last frame.
    wheel_dY = wheelPosY - last_wheelPosY
    last_wheelPosY = wheelPosY

    # Change the orientation of the visual grating based on any vertical mouse wheel movement.
    grating.setOri(wheel_dY * 5, '+')

    # Advance 0.05 cycles per frame.
    grating.setPhase(0.05, '+')

    # Redraw the stim for this frame.
    fixSpot.draw()
    grating.draw()
    message.draw()
    message3.draw()
    displayIdMsg.draw()
    flip_time = win.flip()  # redraw the buffer

    # Check for keyboard and mouse events.
    # If 15 seconds passes without receiving any mouse events, then exit the demo
    kb_events = keyboard.getEvents()
    mouse_events = mouse.getEvents()
    if mouse_events:
        demo_timeout_start = mouse_events[-1].time

    if flip_time - demo_timeout_start > 15.0:
        print("Ending Demo Due to 15 Seconds of Inactivity.")
        break

    # Clear out events that were not accessed this frame.
    io.clearEvents()

# End of Example

win.close()
core.quit()

# The contents of this file are in the public domain.



from .shape import ShapeStim
from .basevisual import ColorMixin, WindowMixin
from psychopy.colors import Color
from .. import layout

knownStyles = ["circles", "cross", ]


class TargetStim(ColorMixin, WindowMixin):
    """
    A target for use in eyetracker calibration, if converted to a dict will return in the correct format for ioHub
    """
    def __init__(self,
                 win, name=None, style="circles",
                 radius=.05, fillColor=(1, 1, 1, 0.1), borderColor="white", lineWidth=2,
                 innerRadius=.01, innerFillColor="red", innerBorderColor=None, innerLineWidth=None,
                 pos=(0, 0), units=None, anchor="center",
                 colorSpace="rgb",
                 autoLog=None, autoDraw=False):
        self.win = win
        # Make sure name is a string
        if name is None:
            name = "target"
        # Create shapes
        self.outer = ShapeStim(win, name=name,
                               vertices="circle",
                               size=(0, 0), pos=pos,
                               lineWidth=lineWidth, units=units,
                               fillColor=fillColor, lineColor=borderColor, colorSpace=colorSpace,
                               autoLog=autoLog, autoDraw=autoDraw)
        self.outerRadius = radius
        self.inner = ShapeStim(win, name=name+"Inner",
                               vertices="circle",
                               size=(0, 0), pos=pos, units=units,
                               lineWidth=(innerLineWidth or lineWidth),
                               fillColor=innerFillColor, lineColor=innerBorderColor, colorSpace=colorSpace,
                               autoLog=autoLog, autoDraw=autoDraw)
        self.innerRadius = innerRadius

        self.style = style
        self.anchor = anchor

    @property
    def style(self):
        if hasattr(self, "_style"):
            return self._style

    @style.setter
    def style(self, value):
        self._style = value
        if value == "circles":
            # Two circles
            self.outer.vertices = self.inner.vertices = "circle"
        elif value == "cross":
            # Circle with a cross inside
            self.outer.vertices = "circle"
            self.inner.vertices = "cross"

    @property
    def scale(self):
        # Work out current ratio between inner and outer radiuses
        return self.innerRadius / self.outerRadius

    @scale.setter
    def scale(self, value):
        self.innerRadius = self.outerRadius * value

    @property
    def anchor(self):
        return self.outer.anchor

    @anchor.setter
    def anchor(self, value):
        self.outer.anchor = value
        self.inner.pos = self.pos + self.size * self.outer._vertices.anchorAdjust

    @property
    def pos(self):
        """For target stims, pos is overloaded so that it moves both the inner and outer shapes."""
        return self.outer.pos

    @pos.setter
    def pos(self, value):
        # Do base pos setting
        WindowMixin.size.fset(self, value)

        self.outer.pos = value
        self.inner.pos = value

    def setPos(self, value):
        self.pos = value

    @property
    def size(self):
        return self.outer.size

    @size.setter
    def size(self, value):
        # Do base size setting
        WindowMixin.size.fset(self, value)
        # Get original scale
        ogScale = self.scale
        # Set new sizes
        self.outer.size = value
        self.inner.size = self.outer.size * ogScale

    @property
    def units(self):
        if hasattr(self, "outer"):
            return self.outer.units

    @units.setter
    def units(self, value):
        if hasattr(self, "outer"):
            self.outer.units = value
        if hasattr(self, "inner"):
            self.inner.units = value

    @property
    def win(self):
        return WindowMixin.win.fget(self)

    @win.setter
    def win(self, value):
        WindowMixin.win.fset(self, value)
        if hasattr(self, "inner"):
            self.inner.win = value
        if hasattr(self, "outer"):
            self.outer.win = value

    @property
    def lineWidth(self):
        return self.outer.lineWidth

    @lineWidth.setter
    def lineWidth(self, value):
        self.outer.lineWidth = value

    @property
    def radius(self):
        return self.outerRadius

    @radius.setter
    def radius(self, value):
        # Set outer radius
        self.outerRadius = value
        # Set inner radius to maintain scale
        self.innerRadius = value * self.scale

    @property
    def outerRadius(self):
        return self.outer.size[1]/2

    @outerRadius.setter
    def outerRadius(self, value):
        # Make buffer object to handle unit conversion
        _buffer = layout.Size((0, value * 2), units=self.units, win=self.win)
        # Use height of buffer object twice, so that size is always square even in norm
        self.outer.size = layout.Size((_buffer.pix[1], _buffer.pix[1]), units='pix', win=self.win)

    @property
    def innerRadius(self):
        return self.inner.size[1] / 2

    @innerRadius.setter
    def innerRadius(self, value):
        # Make buffer object to handle unit conversion
        _buffer = layout.Size((0, value * 2), units=self.units, win=self.win)
        # Use height of buffer object twice, so that size is always square even in norm
        self.inner.size = layout.Size((_buffer.pix[1], _buffer.pix[1]), units='pix', win=self.win)

    @property
    def foreColor(self):
        # Return whichever inner color is not None
        if self.inner.fillColor is not None:
            return self.inner.fillColor
        if self.inner.borderColor is not None:
            return self.inner.borderColor

    @foreColor.setter
    def foreColor(self, value):
        ColorMixin.foreColor.fset(self, value)
        # Set whichever inner color is not None
        if self.inner.fillColor is not None:
            self.inner.fillColor = value
        if self.inner.borderColor is not None:
            self.inner.borderColor = value

    @property
    def borderColor(self):
        return self.outer.borderColor

    @borderColor.setter
    def borderColor(self, value):
        ColorMixin.borderColor.fset(self, value)
        self.outer.borderColor = value

    @property
    def fillColor(self):
        return self.outer.fillColor

    @fillColor.setter
    def fillColor(self, value):
        ColorMixin.fillColor.fset(self, value)
        self.outer.fillColor = value

    @property
    def colorSpace(self):
        return self.outer.colorSpace

    @colorSpace.setter
    def colorSpace(self, value):
        self.outer.colorSpace = value
        self.inner.colorSpace = value

    @property
    def opacity(self):
        return self.outer.opacity

    @opacity.setter
    def opacity(self, value):
        self.outer.opacity = value
        self.inner.opacity = value

    def draw(self, win=None, keepMatrix=False):
        self.outer.draw(win, keepMatrix)
        self.inner.draw(win, keepMatrix)

    def __iter__(self):
        """Overload dict() method to return in ioHub format"""
        # ioHub doesn't treat None as transparent, so we need to handle transparency here
        # For outer circle, use window color as transparent
        fillColor = self.outer.fillColor if self.outer._fillColor else self.win.color
        borderColor = self.outer.borderColor if self.outer._borderColor else self.win.color
        # For inner circle, use outer circle fill as transparent
        innerFillColor = self.inner.fillColor if self.inner._fillColor else fillColor
        innerBorderColor = self.inner.borderColor if self.inner._borderColor else borderColor
        # Assemble dict
        asDict = {
            # Outer circle
            'outer_diameter': self.radius * 2,
            'outer_stroke_width': self.outer.lineWidth,
            'outer_fill_color': fillColor,
            'outer_line_color': borderColor,
            # Inner circle
            'inner_diameter': self.innerRadius * 2,
            'inner_stroke_width': self.inner.lineWidth,
            'inner_fill_color': innerFillColor,
            'inner_line_color': innerBorderColor,
        }
        for key, value in asDict.items():
            yield key, value


def targetFromDict(win, spec,
                   name="target", style="circles",
                   pos=(0, 0), units='height',
                   colorSpace="rgb",
                   autoLog=None, autoDraw=False):
    # Make sure spec has all the required keys, even if it just fills them with None
    required = [
        'outer_diameter', 'outer_stroke_width', 'outer_fill_color', 'outer_line_color',
        'inner_diameter', 'inner_stroke_width', 'inner_fill_color', 'inner_line_color'
    ]
    for key in required:
        if key not in spec:
            spec[key] = None
    # Make a target stim from spec
    TargetStim(win, name=name, style=style,
               radius=spec['outer_diameter']/2, lineWidth=spec['outer_stroke_width'],
               fillColor=spec['outer_fill_color'], borderColor=spec['outer_line_color'],
               innerRadius=spec['outer_diameter']/2, innerLineWidth=spec['inner_stroke_width'],
               innerFillColor=spec['inner_fill_color'], innerBorderColor=spec['inner_line_color'],
               pos=pos, units=units,
               colorSpace=colorSpace,
               autoLog=autoLog, autoDraw=autoDraw)


import os
import subprocess
import sys
import builtins

import wx
import wx.lib.agw.aui as aui
import wx.stc as stc
import wx.richtext
from psychopy.localization import _translate
from wx import py
import numpy
import keyword
from pathlib import Path
from psychopy import prefs
from psychopy import logging
import psychopy
from ...experiment import components
import json

if sys.platform=='win32':
    from matplotlib import font_manager
    fm = font_manager.FontManager()

thisFolder = Path(__file__).parent
iconsPath = Path(prefs.paths['resources'])

try:
    FileNotFoundError
except NameError:
    # Py2 has no FileNotFoundError
    FileNotFoundError = IOError

allCompons = components.getAllComponents()  # ensures that the icons get checked

# Create library of "on brand" colours
cLib = {
    'none': [127, 127, 127, 0],
    'black': [0, 0, 0],
    'grey': [102, 102, 110],
    'white': [242, 242, 242],
    'red': [242, 84, 91],
    'green': [108, 204, 116],
    'blue': [2, 169, 234],
    'yellow': [241, 211, 2],
    'orange': [236, 151, 3],
    'purple': [195, 190, 247],
    'darker': {},
    'lighter': {},
    'very': {'lighter': {},
             'darker': {}}
}
# Create light and dark variants of each colour by +-15 to each value
for c in cLib:
     if not c in ['darker', 'lighter', 'none', 'very']:
         cLib['darker'][c] = [max(0, n-15) for n in cLib[c]]
         cLib['lighter'][c] = [min(255, n+15) for n in cLib[c]]
# Create very light and very dark variants of each colour by a further +-30 to each value
for c in cLib['lighter']:
    cLib['very']['lighter'][c] = [min(255, n+30) for n in cLib['lighter'][c]]
for c in cLib['darker']:
    cLib['very']['darker'][c] = [max(0, n-30) for n in cLib['darker'][c]]


class ThemeMixin:
    lexers = {
        stc.STC_LEX_PYTHON: "python",
        stc.STC_LEX_CPP: "c++",
        stc.STC_LEX_R: "R"
        #stc.STC_LEX_JSON: "json"
    }
    # these are populated and modified by PsychoPyApp.theme.setter
    spec = None
    codetheme = 'PsychopyLight'
    mode = 'light'
    icons = 'light'
    codeColors = {}
    appColors = {}
    appIcons = {'components': {},
                'resources': {}}

    def loadThemeSpec(self, themeName):
        """Load a spec file from disk"""
        # a theme spec contains the spec for the *code* theme as well as a mode
        # that determines which colorscheme to load for the app (separate)
        themesPath = Path(prefs.paths['themes'])

        # first load the *theme* which contains the mode name for the app
        try:
            with open(str(themesPath / (themeName+".json")), "rb") as fp:
                ThemeMixin.spec = themeSpec = json.load(fp)
        except FileNotFoundError:
            with open(str(themesPath / "PsychopyLight.json"), "rb") as fp:
                ThemeMixin.spec = themeSpec = json.load(fp)
        appColorMode = themeSpec['app']
        # Get app spec
        try:
            with open(str(themesPath / "app/{}.json".format(appColorMode)), "rb") as fp:
                ThemeMixin.spec = appColors = json.load(fp)
        except FileNotFoundError:
            with open(str(themesPath / "app/light.json"), "rb") as fp:
                ThemeMixin.spec = appColors = json.load(fp)

        # Set app theme
        ThemeMixin.mode = appColorMode
        self._setAppColors(appColors)
        # Set app icons
        if 'icons' in themeSpec:
            ThemeMixin.icons = themeSpec['icons']
        else:
            ThemeMixin.icons = themeSpec['app']
        # Set coder theme
        codertheme = themeSpec
        ThemeMixin.codetheme = themeName
        self._setCodeColors(codertheme)

    def _applyAppTheme(self, target=None):
        """Applies colorScheme recursively to the target and its children

        Parameters
        ----------
        colorScheme: the new color spec being applied (dict)
        target: the wx object to which being applied
        depth: depth in the tree of wx objects

        """

        # Define subfunctions to handle different object types
        def applyToToolbar(target):
            target.SetBackgroundColour(ThemeMixin.appColors['frame_bg'])
            # Clear tools
            target.ClearTools()
            # Redraw tools
            target.makeTools()

        def applyToStatusBar(target):
            target.SetBackgroundColour(cLib['white'])

        def applyToFrame(target):
            target.SetBackgroundColour(ThemeMixin.appColors['frame_bg'])
            target.SetForegroundColour(ThemeMixin.appColors['text'])
            if hasattr(target, 'GetAuiManager'):
                target.GetAuiManager().SetArtProvider(PsychopyDockArt())
                target.GetAuiManager().Update()
            for menu in target.GetMenuBar().GetMenus():
                for submenu in menu[0].MenuItems:
                    if isinstance(submenu.SubMenu, ThemeSwitcher):
                        submenu.SubMenu._applyAppTheme()


        def applyToPanel(target):
            target.SetBackgroundColour(ThemeMixin.appColors['panel_bg'])
            target.SetForegroundColour(ThemeMixin.appColors['text'])

        def applyToNotebook(target):
            # Dict of icons to apply to specific tabs
            tabIcons = {
                "Structure": "coderclass16.png",
                "FileBrowser": "folder-open16.png",
                "PythonShell": "coderpython16.png",
                "ConsoleOutput": "stdout.png",
            }
            target.SetArtProvider(PsychopyTabArt())
            target.GetAuiManager().SetArtProvider(PsychopyDockArt())
            for index in range(target.GetPageCount()):
                page = target.GetPage(index)
                page.SetBackgroundColour(ThemeMixin.appColors['panel_bg'])
                if page.GetName() in tabIcons:
                    bmp = IconCache.getBitmap(IconCache(), tabIcons[page.GetName()])
                    target.SetPageBitmap(index, bmp)
                page._applyAppTheme()

        def applyToCodeEditor(target):
            spec = ThemeMixin.codeColors.copy()
            base = spec['base']
            # Set margin size according to text size
            if not isinstance(target, wx.py.shell.Shell):
                target.SetMarginWidth(0, 4 * prefs.coder['codeFontSize'])
            # Override base font with user spec if present
            prefkey = 'outputFont' if isinstance(target, wx.py.shell.Shell) else 'codeFont'
            if prefs.coder[prefkey].lower() != "From Theme...".lower():
                for key in spec:
                    if 'font' in spec[key]:
                        spec[key]['font'] = prefs.coder[prefkey] if spec[key]['font'] == base['font'] \
                            else base['font']
                base['font'] = prefs.coder[prefkey]

            # Check that key is in tag list
            invalid = []
            for key in spec:
                if key not in self.tags:
                    invalid += [key]
            for key in invalid:
                del spec[key]

            # Check for language specific spec
            if target.GetLexer() in target.lexers:
                lexer = target.lexers[target.GetLexer()]
            else:
                lexer = 'invlex'
            if lexer in spec:
                # If there is lang specific spec, delete subkey...
                lang = spec['lexer'] # ...and append spec to root, overriding any generic spec
                spec.update({key: lang[key] for key in lang})
            else:
                lang = {}

            # Set style for undefined lexers
            for key in [getattr(wx._stc, item) for item in dir(wx._stc) if item.startswith("STC_LEX")]:
                target.StyleSetBackground(key, base['bg'])
                target.StyleSetForeground(key, base['fg'])
                target.StyleSetSpec(key, "face:%(font)s,size:%(size)d" % base)
            # Set style from universal data
            for key in spec:
                if target.tags[key] is not None:
                    target.StyleSetBackground(target.tags[key], spec[key]['bg'])
                    target.StyleSetForeground(target.tags[key], spec[key]['fg'])
                    target.StyleSetSpec(target.tags[key], "face:%(font)s,size:%(size)d" % spec[key])
            # Apply keywords
            for level, val in target.lexkw.items():
                target.SetKeyWords(level, " ".join(val))

            # Set margin
            target.SetFoldMarginColour(True, spec['margin']['bg'])
            target.SetFoldMarginHiColour(True, spec['margin']['bg'])
            # Set caret colour
            target.SetCaretForeground(spec['caret']['fg'])
            target.SetCaretLineBackground(spec['caret']['bg'])
            target.SetCaretWidth(1 + ('bold' in spec['caret']['font']))
            # Set selection colour
            target.SetSelForeground(True, spec['select']['fg'])
            target.SetSelBackground(True, spec['select']['bg'])
            # Set wrap point
            target.edgeGuideColumn = target.prefs['edgeGuideColumn']
            target.edgeGuideVisible = target.edgeGuideColumn > 0
            # Set line spacing
            spacing = min(int(target.prefs['lineSpacing'] / 2), 64)  # Max out at 64
            target.SetExtraAscent(spacing)
            target.SetExtraDescent(spacing)

        def applyToRichText(target):
            base = ThemeMixin.codeColors['base']
            # todo: Add element-specific styling (it must be possible...)
            # If dealing with a StdOut, set background from base
            target.SetBackgroundColour(
                self.hex2rgb(base['bg'], base['bg']))
            # Then construct default styles
            bold = wx.FONTWEIGHT_BOLD if "bold" in base['font'] else wx.FONTWEIGHT_NORMAL
            italic = wx.FONTSTYLE_ITALIC if "italic" in base['font'] else wx.FONTSTYLE_NORMAL
            # Override base font with user spec if present
            if prefs.coder['outputFont'].lower() == "From Theme...".lower():
                fontName = base['font'].replace("bold", "").replace("italic", "").replace(",", "")
            else:
                fontName = prefs.coder['outputFont']

            _font = wx.Font(
                int(prefs.coder['outputFontSize']),
                wx.FONTFAMILY_TELETYPE, italic,
                bold, False,
                faceName=fontName
            )
            _style = wx.TextAttr(
                colText=wx.Colour(
                    self.hex2rgb(base['fg'], base['fg'])),
                colBack=wx.Colour(
                    self.hex2rgb(base['bg'], base['bg'])),
                font=_font)
            # Then style all text as base
            i = 0
            for ln in range(target.GetNumberOfLines()):
                i += target.GetLineLength(
                    ln) + 1  # +1 as \n is not included in character count
            target.SetDefaultStyle(_style)
            target.SetStyle(0, i, _style)

        def applyToTextCtrl(target):
            base = ThemeMixin.codeColors['base']
            target.SetForegroundColour(base['fg'])
            target.SetBackgroundColour(base['bg'])

        # Define dict linking object types to subfunctions
        handlers = {
            wx.Frame: applyToFrame,
            wx.Panel: applyToPanel,
            aui.AuiNotebook: applyToNotebook,
            psychopy.app.coder.coder.BaseCodeEditor: applyToCodeEditor,
            wx.richtext.RichTextCtrl: applyToRichText,
            wx.py.shell.Shell: applyToCodeEditor,
            wx.ToolBar: applyToToolbar,
            wx.StatusBar: applyToStatusBar,
            wx.TextCtrl: applyToTextCtrl
        }

        # If no target supplied, default to using self
        if target is None:
            target = self

        if not hasattr(self, '_recursionDepth'):
            self._recursionDepth = 0
        else:
            self._recursionDepth += 1

        appCS = ThemeMixin.appColors
        base = ThemeMixin.codeColors['base']
        # Abort if target is immune
        if hasattr(target, 'immune'):
            return

        # Style target
        isHandled = False
        for thisType in handlers:
            if isinstance(target, thisType):
                handlers[thisType](target)
                isHandled = True

        if not isHandled:
            # try and set colors for target
            try:
                target.SetBackgroundColour(ThemeMixin.appColors['panel_bg'])
                target.SetForegroundColour(ThemeMixin.appColors['text'])
            except AttributeError:
                pass

        # search for children (set in a second step)
        if isinstance(target, wx.Sizer):
            sizer = target
            children = sizer.Children
        else:
            children = []
            if hasattr(target, 'Children'):
                children.extend(target.Children)
            elif hasattr(target, 'immune'):
                pass
            elif hasattr(target, 'paneManager'):
                for pane in target.paneManager.AllPanes:
                    children.append(pane.window)
            elif hasattr(target, 'Sizer') and target.Sizer:
                children.append(target.Sizer)

        if hasattr(self, 'btnHandles'):
            for thisBtn in self.btnHandles:
                pass
        # then apply to all children as well
        for c in children:
            if hasattr(c, '_applyAppTheme'):
            # if the object understands themes then request that
                c._applyAppTheme()
            elif self._recursionDepth>10:
                return
            else:
                # if not then use our own recursive method to search
                if hasattr(c, 'Window') and c.Window is not None:
                    ThemeMixin._applyAppTheme(c.Window)
                elif hasattr(c, 'Sizer') and c.Sizer is not None:
                    ThemeMixin._applyAppTheme(c.Sizer)
                # and then apply
                # try:
                #     ThemeMixin._applyAppTheme(c)
                # except AttributeError:
                #     pass

        if hasattr(target, 'Refresh'):
            target.Refresh()
        if hasattr(target, '_mgr'):
            target._mgr.Update()

    @property
    def lexkw(self):
        baseC = {
            0: ['typedef', 'if', 'else', 'return', 'struct', 'for', 'while', 'do',
                'using', 'namespace', 'union', 'break', 'enum', 'new', 'case',
                'switch', 'continue', 'volatile', 'finally', 'throw', 'try',
                'delete', 'typeof', 'sizeof', 'class', 'volatile', 'int',
                'float', 'double', 'char', 'short', 'byte', 'void', 'const',
                'unsigned', 'signed', 'NULL', 'true', 'false', 'bool', 'size_t',
                'long', 'long long'],
            1: []
        }
        if self.GetLexer() == stc.STC_LEX_PYTHON:
            # Python
            keywords = {
                0: keyword.kwlist + ['cdef', 'ctypedef', 'extern', 'cimport', 'cpdef', 'include'],
                1: dir(builtins) + ['self']
            }
        elif self.GetLexer() == stc.STC_LEX_R:
            # R
            keywords = {
                1: ['function', 'for', 'repeat', 'while', 'if', 'else',
                 'break', 'local', 'global'],
                0: ['NA']
            }
        elif self.GetLexer() == stc.STC_LEX_CPP:
            # C/C++
            keywords = baseC.copy()
            if hasattr(self, 'filename'):
                if self.filename.endswith('.js'):
                    # JavaScript
                    keywords = {
                        0: ['var', 'const', 'let', 'import', 'function', 'if',
                            'else', 'return', 'struct', 'for', 'while', 'do',
                            'finally', 'throw', 'try', 'switch', 'case',
                            'break'],
                        1: ['null', 'false', 'true']
                    }
                elif any([self.filename.lower().endswith(ext) for ext in (
                        '.glsl', '.vert', '.frag')]):
                    # keywords
                    keywords[0] += [
                        'invariant', 'precision', 'highp', 'mediump', 'lowp',
                        'coherent', 'sampler', 'sampler2D', 'layout', 'out',
                        'in', 'varying', 'uniform', 'attribute']
                    # types
                    keywords[0] += [
                        'vec2', 'vec3', 'vec4', 'mat2', 'mat3', 'mat4',
                        'ivec2', 'ivec3', 'ivec4', 'imat2', 'imat3', 'imat4',
                        'bvec2', 'bvec3', 'bvec4', 'bmat2', 'bmat3', 'bmat4',
                        'dvec2', 'dvec3', 'dvec4', 'dmat2', 'dmat3', 'dmat4']
                    # reserved
                    keywords[1] += [
                        'gl_Position', 'gl_LightSourceParameters',
                        'gl_MaterialParameters', 'gl_LightModelProducts',
                        'gl_FrontLightProduct', 'gl_BackLightProduct',
                        'gl_FrontMaterial', 'gl_BackMaterial', 'gl_FragColor',
                        'gl_ModelViewMatrix', 'gl_ModelViewProjectionMatrix',
                        'gl_Vertex', 'gl_NormalMatrix', 'gl_Normal',
                        'gl_ProjectionMatrix', 'gl_LightSource']

        # elif self.GetLexer() == stc.STC_LEX_ARDUINO:
        #     # Arduino
        #     keywords = {
        #         0: baseC[0],
        #         1: baseC[1] + [
        #             'BIN', 'HEX', 'OCT', 'DEC', 'INPUT', 'OUTPUT', 'HIGH', 'LOW',
        #             'INPUT_PULLUP', 'LED_BUILTIN', 'string', 'array']
        #     }
        # elif self.GetLexer() == stc.STC_LEX_GLSL:
        #     # GLSL
        #     glslTypes = []
        #     baseType = ['', 'i', 'b', 'd']
        #     dim = ['2', '3', '4']
        #     name = ['vec', 'mat']
        #     for i in baseType:
        #         for j in name:
        #             for k in dim:
        #                 glslTypes.append(i + j + k)
        #     keywords = {
        #         0: baseC[0] + ['invariant', 'precision', 'highp', 'mediump', 'lowp', 'coherent',
        #                                 'sampler', 'sampler2D'],
        #         1: baseC[1]
        #     }
        else:
            keywords = {
                0: [],
                1: []
            }
        return keywords

    @property
    def tags(self):
        tags = {
            "base": stc.STC_STYLE_DEFAULT,
            "margin": stc.STC_STYLE_LINENUMBER,
            "caret": None,
            "select": None,
            "indent": stc.STC_STYLE_INDENTGUIDE,
            "brace": stc.STC_STYLE_BRACELIGHT,
            "controlchar": stc.STC_STYLE_CONTROLCHAR
        }
        if self.GetLexer() == stc.STC_LEX_PYTHON:
            # Python
            tags.update({
                "operator": stc.STC_P_OPERATOR,
                "keyword": stc.STC_P_WORD,
                "keyword2": stc.STC_P_WORD2,
                "id": stc.STC_P_IDENTIFIER,
                "num": stc.STC_P_NUMBER,
                "char": stc.STC_P_CHARACTER,
                "str": stc.STC_P_STRING,
                "openstr": stc.STC_P_STRINGEOL,
                "decorator": stc.STC_P_DECORATOR,
                "def": stc.STC_P_DEFNAME,
                "class": stc.STC_P_CLASSNAME,
                "comment": stc.STC_P_COMMENTLINE,
                "commentblock": stc.STC_P_COMMENTBLOCK,
                "documentation": stc.STC_P_TRIPLE,
                "documentation2": stc.STC_P_TRIPLEDOUBLE,
                "whitespace": stc.STC_P_DEFAULT
            })
        elif self.GetLexer() == stc.STC_LEX_R:
            # R
            tags.update({
                "operator": stc.STC_R_OPERATOR,
                "keyword": stc.STC_R_BASEKWORD,
                "keyword2": stc.STC_R_KWORD,
                "id": stc.STC_R_IDENTIFIER,
                "num": stc.STC_R_NUMBER,
                "char": stc.STC_R_STRING2,
                "str": stc.STC_R_STRING,
                "infix": stc.STC_R_INFIX,
                "openinfix": stc.STC_R_INFIXEOL,
                "comment": stc.STC_R_COMMENT,
                "whitespace": stc.STC_R_DEFAULT
            })
        elif self.GetLexer() == stc.STC_LEX_CPP:
            # C/C++
            tags.update({
                "operator": stc.STC_C_OPERATOR,
                "keyword": stc.STC_C_WORD,
                "keyword2": stc.STC_C_WORD2,
                "id": stc.STC_C_IDENTIFIER,
                "num": stc.STC_C_NUMBER,
                "char": stc.STC_C_CHARACTER,
                "str": stc.STC_C_STRING,
                "openstr": stc.STC_C_STRINGEOL,
                "class": stc.STC_C_GLOBALCLASS,
                "comment": stc.STC_C_COMMENT,
                "commentblock": stc.STC_C_COMMENTLINE,
                "commentkw": stc.STC_C_COMMENTDOCKEYWORD,
                "commenterror": stc.STC_C_COMMENTDOCKEYWORDERROR,
                "documentation": stc.STC_C_COMMENTLINEDOC,
                "documentation2": stc.STC_C_COMMENTDOC,
                "whitespace": stc.STC_C_DEFAULT,
                "preprocessor": stc.STC_C_PREPROCESSOR,
                "preprocessorcomment": stc.STC_C_PREPROCESSORCOMMENT
            })
        # elif self.GetLexer() == stc.STC_LEX_JSON:
        #     # JSON
        #     tags.update({
        #         "operator": stc.STC_JSON_OPERATOR,
        #         "keyword": stc.STC_JSON_KEYWORD,
        #         "uri": stc.STC_JSON_URI,
        #         "compactiri": stc.STC_JSON_COMPACTIRI,
        #         "error": stc.STC_JSON_ERROR,
        #         "espacesequence": stc.STC_JSON_ESCAPESEQUENCE,
        #         "propertyname": stc.STC_JSON_PROPERTYNAME,
        #         "ldkeyword": stc.STC_JSON_LDKEYWORD,
        #         "num": stc.STC_JSON_NUMBER,
        #         "str": stc.STC_JSON_STRING,
        #         "openstr": stc.STC_JSON_STRINGEOL,
        #         "comment": stc.STC_JSON_LINECOMMENT,
        #         "commentblock": stc.STC_JSON_BLOCKCOMMENT,
        #         "whitespace": stc.STC_JSON_DEFAULT
        #     })
        return tags

    def hex2rgb(self, hex, base=(0, 0, 0, 255)):
        if not isinstance(hex, str):
            return base
        # Make hex code case irrelevant
        hex = hex.lower()
        # dict of hex -> int conversions
        hexkeys = {'0': 0, '1': 1, '2': 2, '3': 3, '4': 4, '5': 5, '6': 6, '7': 7, '8': 8, '9': 9,
                   'a': 10, 'b': 11, 'c': 12, 'd': 13, 'e': 14, 'f': 15,
                   '#': None}
        # Check that hex is a hex code
        if not all(c in hexkeys.keys() for c in hex) or not len(hex) == 7:
            # Default to transparent if not
            return wx.Colour(base)

        # Convert to rgb
        r = hexkeys[hex[1]] * 16 + hexkeys[hex[2]]
        g = hexkeys[hex[3]] * 16 + hexkeys[hex[4]]
        b = hexkeys[hex[5]] * 16 + hexkeys[hex[6]]
        return wx.Colour(r, g, b, 255)

    def shiftColour(self, col, offset=15):
        """Shift colour up or down by a set amount"""
        if not isinstance(col, wx.Colour):
            return
        if col.GetLuminance() < 0.5:
            newCol = wx.Colour(
                [c+offset for c in col.Get()]
            )
        else:
            newCol = wx.Colour(
                [c - offset for c in col.Get()]
            )

        return newCol

    def extractFont(self, fontList, base=[]):
        """Extract specified font from theme spec"""
        # Convert to list if not already
        if isinstance(base, str):
            base = base.split(",")
            base = base if isinstance(base, list) else [base]
        if isinstance(fontList, str):
            fontList = fontList.split(",")
            fontList = fontList if isinstance(fontList, list) else [fontList]
        # Extract styles
        bold, italic = [], []
        if "bold" in fontList:
            bold = [fontList.pop(fontList.index("bold"))]
        if "italic" in fontList:
            italic = [fontList.pop(fontList.index("italic"))]
        # Extract styles from base, if needed
        if "bold" in base:
            bold = [base.pop(base.index("bold"))]
        if "italic" in base:
            italic = [base.pop(base.index("italic"))]
        # Append base and default fonts
        fontList.extend(base+["JetBrains Mono"])
        if "" in fontList:
            del fontList[fontList.index("")]
        # Set starting font in case none are found
        try:
            finalFont = [wx.SystemSettings.GetFont(wx.SYS_ANSI_FIXED_FONT).GetFaceName()]
        except wx._core.wxAssertionError:
            finalFont = [wx.Font(10, wx.FONTFAMILY_TELETYPE, wx.FONTSTYLE_NORMAL, wx.FONTWEIGHT_NORMAL).GetFaceName()]
        # Cycle through font names, stop at first valid font
        if sys.platform == 'win32':
            for font in fontList:
                if font in wx.FontEnumerator().GetFacenames():
                    finalFont = [font] + bold + italic
                    break
        return ','.join(finalFont)

    def _setCodeColors(self, spec):
        """To be called from _psychopyApp only"""
        #if not self.GetTopWindow() == self:
        #    psychopy.logging.warning("This function should only be called from _psychopyApp")

        base = spec['base']
        base['font'] = self.extractFont(base['font'])

        # Make sure there's some spec for margins
        if 'margin' not in spec:
            spec['margin'] = base
        # Make sure there's some spec for caret
        if 'caret' not in spec:
            spec['caret'] = base
        # Make sure there's some spec for selection
        if 'select' not in spec:
            spec['select'] = base
            spec['select']['bg'] = self.shiftColour(base['bg'], 30)

        # Pythonise the universal data (hex -> rgb, tag -> wx int)
        invalid = []
        for key in spec:
            # Check that full spec is defined, discard if not
            if all(subkey in spec[key] for subkey in ['bg', 'fg', 'font']):
                spec[key]['bg'] = self.hex2rgb(spec[key]['bg'], base['bg'])
                spec[key]['fg'] = self.hex2rgb(spec[key]['fg'], base['fg'])
                spec[key]['font'] = self.extractFont(spec[key]['font'], base['font'])
                spec[key]['size'] = int(prefs.coder['codeFontSize'])
            elif key in ['app', 'icons']:
                pass
            else:
                invalid += [key]
        for key in invalid:
            del spec[key]

        # we have a valid theme so continue
        for key in spec:
            ThemeMixin.codeColors[key] = spec[key]  # class attribute for all mixin subclasses

    def _setAppColors(self, spec):

        hexchars = ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9',
                    'a', 'b', 'c', 'd', 'e', 'f']
        formats = {
            "hex|named": [str],
            "subnamed1": [str, str],
            "subnamed2": [str, str, str],
            "hex|named_opacity1": [str, int],
            "subnamed1_opacity1": [str, str, int],
            "subnamed2_opacity1": [str, str, str, int],
            "hex|named_opacity2": [str, float],
            "subnamed1_opacity2": [str, str, float],
            "subnamed2_opacity2": [str, str, str, float]
        }
        # Cycle through all values
        for key in spec:
            # if key not in output:
            #    continue
            val = spec[key]
            color = ['invalid']
            # Make sure every value is a list
            if not isinstance(val, list):
                val = [val]

            # Figure out what format current spec is in
            types = [type(v) for v in val]
            format = "invalid"
            for f in formats:
                if formats[f] == types:
                    format = f
            # Pop out opacity so that it can be assumed not present
            if "_opacity" in format:
                opacity = round(val.pop(-1))
                format = format.replace("_opacity", "")
            else:
                opacity = 255
            # Tell the difference between hex and single named values
            if "hex|named" in format:
                if val[0] in cLib:
                    # Extract named colour
                    color = cLib[val[0]]
                    format = format.replace("hex|", "")
                elif len(val[0]) == 7:
                    hex = val[0]
                    if hex[0] == "#" and all([h in hexchars for h in hex[1:].lower()]):
                        # Convert hex colour
                        format = format.replace("|named", "")
                        wxcolor = ThemeMixin.hex2rgb(None, hex)
                        color = list(wxcolor[:3])
                    else:
                        format = "invalid"
                else:
                    format = "invalid"

            if "subnamed" in format:
                if len(val) == 2 and all([v in cLib for v in val]):
                    color = cLib[val[0]][val[1]]
                elif len(val) == 3 and all([v in cLib for v in val]):
                    color = cLib[val[0]][val[1]][val[2]]
                else:
                    format = "invalid"

            if format == "invalid" \
                    or "color" not in locals() \
                    or "opacity" not in locals() \
                    or "invalid" in color:
                raise Exception("Invalid app colour spec")
            else:
                ThemeMixin.appColors[key] = wx.Colour(color + [opacity])


def getBitmap(name, theme, size=None,
                emblem=None, emblemPos='bottom_right'):
    """Retrieves the wx.Bitmap based on name, theme, size and emblem"""
    global _allIcons
    return _allIcons.getBitmap(name, theme, size, emblem, emblemPos)


class IconCache:
    """A class to load icons and store them just once as a dict of wx.Bitmap
    objects according to theme"""
    _theme = ThemeMixin
    _bitmaps = {}
    _buttons = []  # a list of all created buttons
    _lastBGColor = None
    _lastIcons = None

    # def _loadComponentIcons(self, folderList=(), theme=None, forceReload=False):
    #     """load the icons for all the components
    #     """
    #     if theme is None:
    #         theme = _IconCache.iconTheme
    #     if forceReload or len(self)==0:
    #         compons = experiment.getAllComponents(folderList)
    #         _allIcons = {}
    #         for thisName, thisCompon in compons.items():
    #             if thisName in components.iconFiles:
    #                 # darkmode paths
    #                 if "base.png" not in components.iconFiles[thisName]:
    #                     iconFolder = theme
    #                     components.iconFiles[thisName] = join(
    #                             dirname(components.iconFiles[thisName]),
    #                             iconFolder,
    #                             basename(components.iconFiles[thisName])
    #                     )
    #                 _allIcons[thisName] = self._loadIcons(
    #                         components.iconFiles[thisName])
    #             else:
    #                 _allIcons[thisName] = self._loadIcons(None)
    #         return _allIcons
    #     else:
    #         return _allIcons

    def _findImageFile(self, name, theme, emblem=None, size=None):
        """Tries to find a valid icon in a range of places with and without a
        size suffix"""
        orig = Path(name)
        if not orig.suffix:  # check we have an image suffix
            orig = Path(name+'.png')
        if emblem:  # add the emblem into the name
            orig = orig.with_name(
                    "{}_{}{}".format(orig.stem, emblem, orig.suffix))
        nameAndSize = orig.with_name(orig.stem+str(size)+orig.suffix)
        nameAndDouble = orig.with_name(orig.stem+str(size)+"@2x"+orig.suffix)
        for filename in [nameAndSize, orig, nameAndDouble]:
            # components with no themes folders (themes were added in 2020.2)
            if filename.exists():
                return str(filename)
            # components with theme folders
            # try using the theme name (or 'light' as a default name)
            for themeName in [theme, 'light']:
                thisPath = filename.parent / themeName / filename.name
                if thisPath.exists():
                    return str(thisPath)
            # try in the app icons folder (e.g. for "run.png")
            thisPath = iconsPath / theme / filename
            if thisPath.exists():
                return str(thisPath)
            # and in the root of the app icons
            thisPath = iconsPath / filename
            if thisPath.exists():
                return str(thisPath)
        # still haven't returned nay path. Out of ideas!
        logging.warning("Failed to find icon name={}, theme={}, "
                        "size={}, emblem={}"
                        .format(name, theme, size, emblem))

    def _loadBitmap(self, name, theme, size=None, emblem=None):
        """Creates wxBitmaps based on the image.
        png files work best, but anything that wx.Image can load should be fine

        Doesn't return the icons, just stores them in the dict
        """
        filename = self._findImageFile(name, theme, emblem, size)
        if not filename:
            filename = self._findImageFile('unknown.png', theme, emblem, size)

        # load image with wx.LogNull() to stop libpng complaining about sRGB
        nologging = wx.LogNull()
        try:
            im = wx.Image(filename)
        except TypeError:
            raise FileNotFoundError("Failed to find icon name={}, theme={}, "
                        "size={}, emblem={}"
                        .format(name, theme, size, emblem))
        del nologging  # turns logging back on

        pix = im.GetSize()[0]
        if pix > size:
            im = im.Scale(pix, pix)
        nameMain = _getIdentifier(name, theme, emblem, size)
        self._bitmaps[nameMain] = wx.Bitmap(im)
        if pix > 24:  # for bigger images lets create a 1/2 size one too
            nameSmall = _getIdentifier(name, theme, emblem, pix//2)
            self._bitmaps[nameSmall] = wx.Bitmap(im.Scale(pix//2, pix//2))

    def getBitmap(self, name, theme=None, size=None, emblem=None, beta=False):
        """Retrieves an icon based on its name, theme, size and emblem
        either from the cache or loading from file as needed"""
        if theme is None:
            theme = ThemeMixin.icons
        if size is None:
            size = 48
        identifier = _getIdentifier(name, theme=theme, emblem=emblem, size=size)
        # find/load the bitmaps first
        if identifier not in IconCache._bitmaps:
            # load all size icons for this name
            self._loadBitmap(name, theme, emblem=emblem, size=size)

        if beta:
            # If needed, append beta tag
            betaID = _getIdentifier("beta", theme=theme, emblem=emblem, size=size)
            if betaID not in IconCache._bitmaps:
                self._loadBitmap("beta", theme, emblem=emblem, size=size)
            # Get base icon and beta overlay
            betaImg = IconCache._bitmaps[betaID].ConvertToImage()
            baseImg = IconCache._bitmaps[identifier].ConvertToImage()
            # Get color data and alphas
            betaData = numpy.array(betaImg.GetData())
            betaAlpha = numpy.array(betaImg.GetAlpha(), dtype=int)
            baseData = numpy.array(baseImg.GetData())
            baseAlpha = numpy.array(baseImg.GetAlpha(), dtype=int)
            # Overlay colors
            combinedData = baseData
            r = numpy.where(betaAlpha > 0)[0] * 3
            g = numpy.where(betaAlpha > 0)[0] * 3 + 1
            b = numpy.where(betaAlpha > 0)[0] * 3 + 2
            combinedData[r] = betaData[r]
            combinedData[g] = betaData[g]
            combinedData[b] = betaData[b]
            # Combine alphas
            combinedAlpha = numpy.add(baseAlpha, betaAlpha)
            combinedAlpha[combinedAlpha > 255] = 255
            combinedAlpha = numpy.uint8(combinedAlpha)
            # Set these back to the base image
            combined = betaImg
            combined.SetData(combinedData)
            combined.SetAlpha(combinedAlpha)
            # Replace icon
            identifier += "_beta"
            IconCache._bitmaps[identifier] = combined.ConvertToBitmap()

        return IconCache._bitmaps[identifier]

    def makeBitmapButton(self, parent, filename,
                         name="",  # name of Component e.g. TextComponent
                         label="", # label on the button, often short name
                         emblem=None,
                         toolbar=None, tip=None, size=None,
                         tbKind=wx.ITEM_NORMAL, theme=None):
        if theme is None:
            theme = ThemeMixin.icons
        bmp = self.getBitmap(filename, theme, size, emblem)
        if toolbar:
            if 'phoenix' in wx.PlatformInfo:
                button = toolbar.AddTool(wx.ID_ANY, label=label,
                                         bitmap=bmp, shortHelp=tip,
                                         kind=tbKind)
            else:
                button = toolbar.AddSimpleTool(wx.ID_ANY, label=label,
                                               bitmap=bmp, shortHelp=tip,
                                               kind=tbKind)
        else:
            button = wx.Button(parent, wx.ID_ANY,
                               label=label, name=name, style=wx.NO_BORDER)
            button.SetBitmap(bmp)
            button.SetBitmapPosition(wx.TOP)
            button.SetBackgroundColour(ThemeMixin.appColors['frame_bg'])
            # just for regular buttons (not toolbar objects) we can re-use
            buttonInfo = {'btn': button,
                          'filename': filename,
                          'size': size,
                          'emblem': emblem,
                          'theme': theme}
            self._buttons.append(buttonInfo)

            if tip:
                button.SetToolTip(wx.ToolTip(tip))

        return button

    def getComponentButton(self, parent, name, label,
                           theme=None, size=None, emblem=None,
                           tip=""):
        """Checks in the experiment.components.iconFiles for filename and
        loads it into a wx.Bitmap"""
        if name in components.iconFiles:
            filename = components.iconFiles[name]
            btn = self.makeBitmapButton(
                    parent=parent,
                    filename=filename, name=name, label=label,
                    tip=tip, size=size)
            return btn

    def getComponentBitmap(self, name, size=None):
        """Checks in the experiment.components.iconFiles for filename and
        loads it into a wx.Bitmap"""
        if type(name) != str:  # got a class instead of a name?
            name = name.getType()
        if name in components.iconFiles:
            filename = components.iconFiles[name]
            bmp = self.getBitmap(name=filename, size=size)
            return bmp
        else:
            print(components.iconFiles)
            raise ValueError("Failed to find '{}' in components.iconFiles"
                             .format(name))

    def setTheme(self, theme):
        if theme.icons != IconCache._lastIcons:
            for thisBtn in IconCache._buttons:
                if thisBtn['btn']:  # Check that button hasn't been deleted
                    newBmp = self.getBitmap(name=thisBtn['filename'],
                                            size=thisBtn['size'],
                                            theme=theme.icons,
                                            emblem=thisBtn['emblem'])
                    thisBtn['btn'].SetBitmap(newBmp)
                    thisBtn['btn'].SetBitmapCurrent(newBmp)
                    thisBtn['btn'].SetBitmapPressed(newBmp)
                    thisBtn['btn'].SetBitmapFocus(newBmp)
                    thisBtn['btn'].SetBitmapDisabled(newBmp)
                    thisBtn['btn'].SetBitmapPosition(wx.TOP)
        IconCache._lastIcons = theme.icons
        if theme.appColors['frame_bg'] != IconCache._lastBGColor:
            for thisBtn in IconCache._buttons:
                try:
                    thisBtn['btn'].SetBackgroundColour(
                            theme.appColors['frame_bg'])
                except RuntimeError:
                    pass
        IconCache._lastBGColor = theme


def _getIdentifier(name, theme, emblem, size=None):
    if size is None:
        return "{}_{}_{}".format(name, theme, emblem)
    else:
        return "{}_{}_{}_{}".format(name, theme, emblem, size)


class PsychopyTabArt(aui.AuiDefaultTabArt, ThemeMixin):
    def __init__(self):
        aui.AuiDefaultTabArt.__init__(self)

        self.SetDefaultColours()
        self.SetAGWFlags(aui.AUI_NB_NO_TAB_FOCUS)

    def SetDefaultColours(self):
        """
        Sets the default colours, which are calculated from the given base colour.

        :param `base_colour`: an instance of :class:`wx.Colour`. If defaulted to ``None``, a colour
         is generated accordingly to the platform and theme.
        """
        cs = ThemeMixin.appColors
        self.SetBaseColour( wx.Colour(cs['tab_bg']) )
        self._background_top_colour = wx.Colour(cs['panel_bg'])
        self._background_bottom_colour = wx.Colour(cs['panel_bg'])

        self._tab_text_colour = lambda page: cs['text']
        self._tab_top_colour = wx.Colour(cs['tab_bg'])
        self._tab_bottom_colour = wx.Colour(cs['tab_bg'])
        self._tab_gradient_highlight_colour = wx.Colour(cs['tab_bg'])
        self._border_colour = wx.Colour(cs['tab_bg'])
        self._border_pen = wx.Pen(self._border_colour)

        self._tab_disabled_text_colour = cs['text']
        self._tab_inactive_top_colour = wx.Colour(cs['panel_bg'])
        self._tab_inactive_bottom_colour = wx.Colour(cs['panel_bg'])

    def DrawTab(self, dc, wnd, page, in_rect, close_button_state, paint_control=False):
        """
        Extends AuiDefaultTabArt.DrawTab to add a transparent border to inactive tabs
        """
        if page.active:
            self._border_pen = wx.Pen(self._border_colour)
        else:
            self._border_pen = wx.TRANSPARENT_PEN

        out_tab_rect, out_button_rect, x_extent = aui.AuiDefaultTabArt.DrawTab(self, dc, wnd, page, in_rect, close_button_state, paint_control)

        return out_tab_rect, out_button_rect, x_extent

class PsychopyDockArt(aui.AuiDefaultDockArt):
    def __init__(self):
        aui.AuiDefaultDockArt.__init__(self)
        cs = ThemeMixin.appColors
        # Gradient
        self._gradient_type = aui.AUI_GRADIENT_NONE
        # Background
        self._background_colour = wx.Colour(cs['frame_bg'])
        self._background_gradient_colour = wx.Colour(cs['frame_bg'])
        self._background_brush = wx.Brush(self._background_colour)
        # Border
        self._border_size = 0
        self._border_pen = wx.Pen(cs['frame_bg'])
        # Sash
        self._draw_sash = True
        self._sash_size = 5
        self._sash_brush = wx.Brush(cs['frame_bg'])
        # Gripper
        self._gripper_brush = wx.Brush(cs['frame_bg'])
        self._gripper_pen1 = wx.Pen(cs['frame_bg'])
        self._gripper_pen2 = wx.Pen(cs['frame_bg'])
        self._gripper_pen3 = wx.Pen(cs['frame_bg'])
        self._gripper_size = 0
        # Hint
        self._hint_background_colour = wx.Colour(cs['frame_bg'])
        # Caption bar
        self._inactive_caption_colour = wx.Colour(cs['docker_bg'])
        self._inactive_caption_gradient_colour = wx.Colour(cs['docker_bg'])
        self._inactive_caption_text_colour = wx.Colour(cs['docker_fg'])
        self._active_caption_colour = wx.Colour(cs['docker_bg'])
        self._active_caption_gradient_colour = wx.Colour(cs['docker_bg'])
        self._active_caption_text_colour = wx.Colour(cs['docker_fg'])
        # self._caption_font
        self._caption_size = 25
        self._button_size = 20


class ThemeSwitcher(wx.Menu):
    """Class to make a submenu for switching theme, meaning that the menu will
    always be the same across frames."""
    def __init__(self, frame):
        # Get list of themes
        themePath = Path(prefs.paths['themes'])
        themeList = {}
        for themeFile in themePath.glob("*.json"):
            try:
                with open(themeFile, "rb") as fp:
                    theme = json.load(fp)
                    # Add themes to list only if min spec is defined
                    base = theme['base']
                    if all(key in base for key in ['bg', 'fg', 'font']):
                            themeList[themeFile.stem] = theme['info'] if "info" in theme else ""

            except (FileNotFoundError, IsADirectoryError):
                pass
        # Make menu
        wx.Menu.__init__(self)
        # Make priority theme buttons
        priority = ["PsychopyDark", "PsychopyLight", "ClassicDark", "Classic"]
        for theme in priority:
            tooltip = themeList.pop(theme)
            item = self.AppendRadioItem(wx.ID_ANY, _translate(theme), tooltip)
            # Bind to theme change method
            frame.Bind(wx.EVT_MENU, frame.app.onThemeChange, item)
        # Make other theme buttons
        for theme in themeList:
            item = self.AppendRadioItem(wx.ID_ANY, _translate(theme), help=themeList[theme])
            frame.Bind(wx.EVT_MENU, frame.app.onThemeChange, item)
        self.AppendSeparator()
        # Add Theme Folder button
        item = self.Append(wx.ID_ANY, _translate("Open theme folder"))
        frame.Bind(wx.EVT_MENU, self.openThemeFolder, item)

    def openThemeFolder(self, event):
        # Choose a command according to OS
        if sys.platform in ['win32']:
            comm = "explorer"
        elif sys.platform in ['darwin']:
            comm = "open"
        elif sys.platform in ['linux', 'linux2']:
            comm = "dolphin"
        # Use command to open themes folder
        subprocess.call(f"{comm} {prefs.paths['themes']}", shell=True)

    def _applyAppTheme(self):
        for item in self.GetMenuItems():
            if item.IsRadio():  # This means it will not attempt to check the separator
                item.Check(item.ItemLabel.lower() == ThemeMixin.codetheme.lower())





import sys
from psychopy import visual, clock
import pytest
import numpy as np

try:
    from psychtoolbox import GetSecs
    havePTB = True
except ImportError:
    havePTB = False

class Test_WinFlipTiming():
    def setup_class(self):
        self.win = visual.Window(size=(200, 200), units='pix',
                                 allowGUI=False, autoLog=False,
                                 waitBlanking=True)
        self.rate = self.win.getActualFrameRate()

        if self.rate and 40<self.rate<160:
            self.testOutputs = True
        else:
            self.testOutputs = False

    def teardown_class(self):
        self.win.close()

    def _runSeriesOfFlips(self, usePTB):
        if usePTB:
            getTime = GetSecs
            clk = 'ptb'
        else:
            getTime = clock.monotonicClock.getTime
            clk = clock.monotonicClock

        self.win.flip()
        now = clock.monotonicClock.getTime()
        next = self.win.getFutureFlipTime(clock=clk)
        self.win.flip()

        errsNext = []
        # check nextFrame against reality for 10 frames
        for frameN in range(10):
            self.win.flip()
            this = getTime()
            # print('this', this)
            # print('err', next-this)
            errsNext.append(next-this)
            #then update next
            next = self.win.getFutureFlipTime(clock=clk)
            # print('next', next)
        return errsNext

    def test_winFutureFlip(self):
        """test window.viewScale and .viewPos simultaneous
        negative-going scale should mirror-reverse, and position should
        account for that visually, the green square/rect should move clockwise
        around the text

        Non-zero viewOri would not currently pass with a nonzero viewPos
        """
        self.win.flip()
        now = clock.monotonicClock.getTime()
        next = self.win._frameTimes[-1] + self.win.monitorFramePeriod

        errs = self._runSeriesOfFlips(usePTB=False)
        print()
        print('getFutureFlipTime(0): mean={:.6f}, std={:6f}, all={}'
              .format(np.mean(errs), np.std(errs), errs))
        if self.testOutputs:
            assert np.mean(errs)<0.005  # checks for any systematic bias

        if havePTB:
            errs = self._runSeriesOfFlips(usePTB=True)
            print('PTB getFutureFlipTime(0): mean={:.6f}, std={:6f}, all={}'
                  .format(np.mean(errs), np.std(errs), errs))
            if self.testOutputs:
                assert np.mean(errs)<0.005  # checks for any systematic bias

        now = clock.monotonicClock.getTime()
        predictedFrames = []
        print('now={:.5f}, lastFrame={:.5f}'.format(now, self.win._frameTimes[-1]))
        print('delay requestT expectT diff'.format(now, self.win._frameTimes[-1]))
        for requested in np.arange(0, 0.04, 0.001):
            expect = self.win.getFutureFlipTime(requested)
            diff = expect-(now+requested)
            print("{:.4f}, {:.5f} {:.5f} {:.5f}".format(requested, requested, expect, diff))
            # should always be within 1/2 frame
        if self.testOutputs and sys.platform != 'darwin':
            assert abs(diff) < self.win.monitorFramePeriod/2.0


if __name__ == "__main__":
    pytest.main(__file__)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Demo of TrialHandler
"""

from random import random
from psychopy import data

# create your list of stimuli
# NB as of version 1.62 you could simply import an excel spreadsheet with this
# using data.importConditions('someFile.xlsx')
stimList = []
for ori in range(90, 180, 30):
    for sf in [0.5, 1.0, 2.0]:
        # append a python 'dictionary' to the list
        stimList.append({'sf':sf, 'ori':ori})

# organize them with the trial handler
trials = data.TrialHandler2(stimList, 10,
                            extraInfo= {'participant':"Nobody", 'session':'001'})

# run the experiment
nDone = 0
for thisTrial in trials:  # handler can act like a for loop
    # simulate some data
    thisReactionTime = random() + float(thisTrial['sf']) / 2.0
    thisChoice = round(random())
    trials.addData('RT', thisReactionTime)  # add the data to our set
    trials.addData('choice', thisChoice)
    nDone += 1  # just for a quick reference

    msg = 'trial %i had position %s in the list (sf=%.1f)'
    print(msg % (nDone, trials.thisIndex, thisTrial['sf']))

# after the experiment
print('\n')
trials.saveAsPickle(fileName = 'testData')  # this saves a copy of the whole object
df = trials.saveAsWideText("testDataWide.csv")  # wide is useful for analysis with R or SPSS. Also returns dataframe df

# The contents of this file are in the public domain.



from pathlib import Path

import pytest

from psychopy import experiment


def _make_minimal_experiment(obj):
    """
    Make a minimal experiment with just one routine, the same class as the current standalone routine but with all
    default params.
    """
    # Skip whole test if required attributes aren't present
    if not hasattr(obj, "rt"):
        pytest.skip()
    # Make blank experiment
    exp = experiment.Experiment()
    # Create instance of this component with all default params
    rtClass = type(obj.rt)
    rt = rtClass(exp=exp, name=f"test{rtClass.__name__}")
    exp.addStandaloneRoutine(rt.name, rt)
    exp.flow.addRoutine(rt, 0)
    # Return experiment, routine and component
    return rt, exp


class _TestBaseStandaloneRoutinesMixin:
    def test_icons(self):
        """Check that routine has icons for each app theme"""
        # Pathify icon file path
        icon = Path(self.rt.iconFile)
        # Get paths for each theme
        files = [
            icon.parent / "light" / icon.name,
            icon.parent / "dark" / icon.name,
            icon.parent / "classic" / icon.name,
        ]
        # Check that each path is a file
        for file in files:
            assert file.is_file()

    def test_params_used(self):
        # Make minimal experiment just for this test
        rt, exp = _make_minimal_experiment(self)
        # Try with PsychoPy and PsychoJS
        for target in ("PsychoPy", "PsychoJS"):
            ## Skip PsychoJS until can write script without saving
            if target == "PsychoJS":
                continue
            # Skip unimplemented targets
            if target not in rt.targets:
                continue
            # Compile script
            script = exp.writeScript(target=target)
            # Check that the string value of each param is present in the script
            experiment.utils.scriptTarget = target
            # Iterate through every param
            for routine in exp.flow:
                for name, param in experiment.getInitVals(routine.params, target).items():
                    # Conditions to skip...
                    if not param.direct:
                        # Marked as not direct
                        continue
                    if any(name in depend['param'] for depend in routine.depends):
                        # Dependent on another param
                        continue
                    # Check that param is used
                    assert str(param) in script, (
                        f"Value {param} of <psychopy.experiment.params.Param: val={param.val}, valType={param.valType}> "
                        f"in {type(rt).__name__} not found in {target} script."
                    )


class _TestDisabledMixin:
    def test_disabled_default_val(self):
        """
        Test that routines created with default params are not disabled
        """
        # Make minimal experiment just for this test
        rt, exp = _make_minimal_experiment(self)
        # Check whether it can be disabled
        assert 'disabled' in rt.params, (
            f"{type(rt).__name__} does not have a 'disabled' attribute."
        )
        # Check that disabled defaults to False
        assert rt.params['disabled'].val is False, f"{type(rt).__name__} is defaulting to disabled."

    def test_code_muting(self):
        """
        Test that routines are only written when enabled and targets match.
        """
        # Make minimal experiment just for this test
        rt, exp = _make_minimal_experiment(self)
        # Write experiment and check that routine is written
        pyScript = exp.writeScript(target="PsychoPy")
        if "PsychoPy" in type(rt).targets:
            assert rt.name in pyScript, (
                f"{type(rt).__name__} not found in compiled Python script when enabled and PsychoPy in targets."
            )
        else:
            assert rt.name not in pyScript, (
                f"{type(rt).__name__} found in compiled Python script when enabled but PsychoPy not in targets."
            )
        # ## disabled until js can compile without saving
        # jsScript = exp.writeScript(target="PsychoJS")
        # if "PsychoJS" in type(rt).targets:
        #     assert rt.name in jsScript, (
        #         f"{type(rt).__name__} not found in compiled Python script when enabled and PsychoJS in targets."
        #     )
        # else:
        #     assert rt.name not in jsScript, (
        #         f"{type(rt).__name__} found in compiled Python script when enabled but PsychoJS not in targets."
        #     )

        # Disable routine then do same tests but assert not present
        rt.params['disabled'].val = True

        pyScript = exp.writeScript(target="PsychoPy")
        if "PsychoPy" in type(rt).targets:
            assert rt.name not in pyScript, (
                f"{type(rt).__name__} found in compiled Python script when disabled but PsychoPy in targets."
            )
        else:
            assert rt.name not in pyScript, (
                f"{type(rt).__name__} found in compiled Python script when disabled and PsychoPy not in targets."
            )
        # ## disabled until js can compile without saving
        # jsScript = exp.writeScript(target="PsychoJS")
        # if "PsychoJS" in type(rt).targets:
        #     assert rt.name not in jsScript, (
        #         f"{type(rt).__name__} found in compiled Python script when disabled but PsychoJS in targets."
        #     )
        # else:
        #     assert rt.name not in jsScript, (
        #         f"{type(rt).__name__} found in compiled Python script when disabled and PsychoJS not in targets."
        #     )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Audio recording using a microphone.
"""

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2022 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

__all__ = ['Microphone']

import sys
import psychopy.logging as logging
from psychopy.constants import NOT_STARTED, STARTED
from psychopy.preferences import prefs
from .audioclip import *
from .audiodevice import *
from .exceptions import *
import numpy as np

_hasPTB = True
try:
    import psychtoolbox.audio as audio
except (ImportError, ModuleNotFoundError):
    logging.warning(
        "The 'psychtoolbox' library cannot be loaded but is required for audio "
        "capture (use `pip install psychtoolbox` to get it). Microphone "
        "recording will be unavailable this session. Note that opening a "
        "microphone stream will raise an error.")
    _hasPTB = False


class RecordingBuffer:
    """Class for a storing a recording from a stream.

    Think of instances of this class behaving like an audio tape whereas the
    `Microphone` class is the tape recorder. Samples taken from the stream are
    written to the tape which stores the data.

    Used internally by the `Microphone` class, users usually do not create
    instances of this class themselves.

    Parameters
    ----------
    sampleRateHz : int
        Sampling rate for audio recording in Hertz (Hz). By default, 48kHz
        (``sampleRateHz=480000``) is used which is adequate for most consumer
        grade microphones (headsets and built-in).
    channels : int
        Number of channels to record samples to `1=Mono` and `2=Stereo`.
    maxRecordingSize : int
        Maximum recording size in kilobytes (Kb). Since audio recordings tend to
        consume a large amount of system memory, one might want to limit the
        size of the recording buffer to ensure that the application does not run
        out of memory. By default, the recording buffer is set to 24000 KB (or
        24 MB). At a sample rate of 48kHz, this will result in 62.5 seconds of
        continuous audio being recorded before the buffer is full.
    policyWhenFull : str
        What to do when the recording buffer is full and cannot accept any more
        samples. If 'ignore', samples will be silently dropped and the `isFull`
        property will be set to `True`. If 'warn', a warning will be logged and
        the `isFull` flag will be set. Finally, if 'error' the application will
        raise an exception.

    """
    def __init__(self, sampleRateHz=SAMPLE_RATE_48kHz, channels=2,
                 maxRecordingSize=24000, policyWhenFull='ignore'):
        self._channels = channels
        self._sampleRateHz = sampleRateHz
        self._maxRecordingSize = maxRecordingSize
        self._samples = None  # `ndarray` created in _allocRecBuffer`
        self._offset = 0  # recording offset
        self._lastSample = 0  # offset of the last sample from stream
        self._spaceRemaining = None  # set in `_allocRecBuffer`
        self._totalSamples = None  # set in `_allocRecBuffer`

        # check if the value is valid
        if policyWhenFull not in ['ignore', 'warn', 'error']:
            raise ValueError("Invalid value for `policyWhenFull`.")

        self._policyWhenFull = policyWhenFull
        self._warnedRecBufferFull = False
        self._loops = 0

        self._allocRecBuffer()

    def _allocRecBuffer(self):
        """Allocate the recording buffer. Called internally if properties are
        changed."""
        # allocate another array
        nBytes = self._maxRecordingSize * 1000
        recArraySize = int((nBytes / self._channels) / (np.float32()).itemsize)

        self._samples = np.zeros(
            (recArraySize, self._channels), dtype=np.float32, order='C')

        # sanity check
        assert self._samples.nbytes == nBytes
        self._totalSamples = len(self._samples)
        self._spaceRemaining = self._totalSamples

    @property
    def samples(self):
        """Reference to the actual sample buffer (`ndarray`)."""
        return self._samples

    @property
    def bufferSecs(self):
        """Capacity of the recording buffer in seconds (`float`)."""
        return self._totalSamples / self._sampleRateHz

    @property
    def nbytes(self):
        """Number of bytes the recording buffer occupies in memory (`int`)."""
        return self._samples.nbytes

    @property
    def sampleBytes(self):
        """Number of bytes per sample (`int`)."""
        return np.float32().itemsize

    @property
    def spaceRemaining(self):
        """The space remaining in the recording buffer (`int`). Indicates the
        number of samples that the buffer can still add before overflowing.
        """
        return self._spaceRemaining

    @property
    def isFull(self):
        """Is the recording buffer full (`bool`)."""
        return self._spaceRemaining <= 0

    @property
    def totalSamples(self):
        """Total number samples the recording buffer can hold (`int`)."""
        return self._totalSamples

    @property
    def writeOffset(self):
        """Index in the sample buffer where new samples will be written when
        `write()` is called (`int`).
        """
        return self._offset

    @property
    def lastSample(self):
        """Index of the last sample recorded (`int`). This can be used to slice
        the recording buffer, only getting data from the beginning to place
        where the last sample was written to.
        """
        return self._lastSample

    @property
    def loopCount(self):
        """Number of times the recording buffer restarted (`int`). Only valid if
        `loopback` is ``True``."""
        return self._loops

    @property
    def maxRecordingSize(self):
        """Maximum recording size in kilobytes (`int`).

        Since audio recordings tend to consume a large amount of system memory,
        one might want to limit the size of the recording buffer to ensure that
        the application does not run out of memory. By default, the recording
        buffer is set to 24000 KB (or 24 MB). At a sample rate of 48kHz, this
        will result in 62.5 seconds of continuous audio being recorded before
        the buffer is full.

        Setting this value will allocate another recording buffer of appropriate
        size. Avoid doing this in any time sensitive parts of your application.

        """
        return self._maxRecordingSize

    @maxRecordingSize.setter
    def maxRecordingSize(self, value):
        value = int(value)

        # don't do this unless the value changed
        if value == self._maxRecordingSize:
            return

        # if different than last value, update the recording buffer
        self._maxRecordingSize = value
        self._allocRecBuffer()

    def seek(self, offset, absolute=False):
        """Set the write offset.

        Use this to specify where to begin writing samples the next time `write`
        is called. You should call `seek(0)` when starting a new recording.

        Parameters
        ----------
        offset : int
            Position in the sample buffer to set.
        absolute : bool
            Use absolute positioning. Use relative positioning if `False` where
            the value of `offset` will be added to the current offset. Default
            is `False`.

        """
        if not absolute:
            self._offset += offset
        else:
            self._offset = absolute

        assert 0 <= self._offset < self._totalSamples
        self._spaceRemaining = self._totalSamples - self._offset

    def write(self, samples):
        """Write samples to the recording buffer.

        Parameters
        ----------
        samples : ArrayLike
            Samples to write to the recording buffer, usually of a stream. Must
            have the same number of dimensions as the internal array.

        Returns
        -------
        int
            Number of samples overflowed. If this is zero then all samples have
            been recorded, if not, the number of samples rejected is given.

        """
        nSamples = len(samples)
        if self.isFull:
            if self._policyWhenFull == 'ignore':
                return nSamples  # samples lost
            elif self._policyWhenFull == 'warn':
                if not self._warnedRecBufferFull:
                    logging.warning(
                        f"Audio recording buffer filled! This means that no "
                        f"samples are saved beyond {round(self.bufferSecs, 6)} "
                        f"seconds. Specify a larger recording buffer next time "
                        f"to avoid data loss.")
                    logging.flush()
                    self._warnedRecBufferFull = True
                return nSamples
            elif self._policyWhenFull == 'error':
                raise AudioRecordingBufferFullError(
                    "Cannot write samples, recording buffer is full.")
            else:
                return nSamples  # whatever

        if not nSamples:  # no samples came out of the stream, just return
            return

        if self._spaceRemaining >= nSamples:
            self._lastSample = self._offset + nSamples
            audioData = samples[:, :]
        else:
            self._lastSample = self._offset + self._spaceRemaining
            audioData = samples[:self._spaceRemaining, :]

        self._samples[self._offset:self._lastSample, :] = audioData
        self._offset += nSamples

        self._spaceRemaining -= nSamples

        # Check if the recording buffer is now full. Next call to `poll` will
        # not record anything.
        if self._spaceRemaining <= 0:
            self._spaceRemaining = 0

        d = nSamples - self._spaceRemaining
        return 0 if d < 0 else d

    def clear(self):
        # reset all live attributes
        self._samples = None
        self._offset = 0
        self._lastSample = 0
        self._spaceRemaining = None
        self._totalSamples = None
        # reallocate buffer
        self._allocRecBuffer()

    def getSegment(self, start=0, end=None):
        """Get a segment of recording data as an `AudioClip`.

        Parameters
        ----------
        start : float or int
            Absolute time in seconds for the start of the clip.
        end : float or int
            Absolute time in seconds for the end of the clip. If `None` the time
            at the last sample is used.

        Returns
        -------
        AudioClip
            Audio clip object with samples between `start` and `end`.

        """
        idxStart = int(start * self._sampleRateHz)
        idxEnd = self._lastSample if end is None else int(
            end * self._sampleRateHz)

        return AudioClip(
            np.array(self._samples[idxStart:idxEnd, :],
                     dtype=np.float32, order='C'),
            sampleRateHz=self._sampleRateHz)


class Microphone:
    """Class for recording audio from a microphone or input stream.

    Creating an instance of this class will open a stream using the specified
    device. Streams should remain open for the duration of your session. When a
    stream is opened, a buffer is allocated to store samples coming off it.
    Samples from the input stream will written to the buffer once
    :meth:`~Microphone.start()` is called.

    Parameters
    ----------
    device : int or `~psychopy.sound.AudioDevice`
        Audio capture device to use. You may specify the device either by index
        (`int`) or descriptor (`AudioDevice`).
    sampleRateHz : int
        Sampling rate for audio recording in Hertz (Hz). By default, 48kHz
        (``sampleRateHz=480000``) is used which is adequate for most consumer
        grade microphones (headsets and built-in).
    channels : int
        Number of channels to record samples to `1=Mono` and `2=Stereo`.
    streamBufferSecs : float
        Stream buffer size to pre-allocate for the specified number of seconds.
        The default is 2.0 seconds which is usually sufficient.
    maxRecordingSize : int
        Maximum recording size in kilobytes (Kb). Since audio recordings tend to
        consume a large amount of system memory, one might want to limit the
        size of the recording buffer to ensure that the application does not run
        out of memory. By default, the recording buffer is set to 24000 KB (or
        24 MB). At a sample rate of 48kHz, this will result in 62.5 seconds of
        continuous audio being recorded before the buffer is full.
    audioLatencyMode : int or None
        Audio latency mode to use, values range between 0-4. If `None`, the
        setting from preferences will be used. Using `3` (exclusive mode) is
        adequate for most applications and required if using WASAPI on Windows
        for other settings (such audio quality) to take effect. Symbolic
        constants `psychopy.sound.audiodevice.AUDIO_PTB_LATENCY_CLASS_` can also
        be used.
    audioRunMode : int
        Run mode for the recording device. Default is standby-mode (`0`) which
        allows the system to put the device to sleep. However when the device is
        needed, waking the device results in some latency. Using a run mode of
        `1` will keep the microphone running (or 'hot') with reduces latency
        when th recording is started. Cannot be set when after initialization at
        this time.

    Examples
    --------
    Capture 10 seconds of audio from the primary microphone::

        import psychopy.core as core
        import psychopy.sound.Microphone as Microphone

        mic = Microphone(bufferSecs=10.0)  # open the microphone
        mic.start()  # start recording
        core.wait(10.0)  # wait 10 seconds
        mic.stop()  # stop recording

        audioClip = mic.getRecording()

        print(audioClip.duration)  # should be ~10 seconds
        audioClip.save('test.wav')  # save the recorded audio as a 'wav' file

    The prescribed method for making long recordings is to poll the stream once
    per frame (or every n-th frame)::

        mic = Microphone(bufferSecs=2.0)
        mic.start()  # start recording

        # main trial drawing loop
        mic.poll()
        win.flip()  # calling the window flip function

        mic.stop()  # stop recording
        audioClip = mic.getRecording()

    """
    # Force the use of WASAPI for audio capture on Windows. If `True`, only
    # WASAPI devices will be returned when calling static method
    # `Microphone.getDevices()`
    enforceWASAPI = True

    def __init__(self,
                 device=None,
                 sampleRateHz=None,
                 channels=2,
                 streamBufferSecs=2.0,
                 maxRecordingSize=24000,
                 policyWhenFull='warn',
                 audioLatencyMode=None,
                 audioRunMode=0):

        if not _hasPTB:  # fail if PTB is not installed
            raise ModuleNotFoundError(
                "Microphone audio capture requires package `psychtoolbox` to "
                "be installed.")

        # get information about the selected device
        devices = Microphone.getDevices()
        if isinstance(device, AudioDeviceInfo):
            self._device = device
        elif isinstance(device, (int, float)):
            devicesByIndex = {d.deviceIndex: d for d in devices}
            if device in devicesByIndex:
                self._device = devicesByIndex[device]
            else:
                raise AudioInvalidCaptureDeviceError(
                    'No suitable audio recording devices found matching index '
                    '{}.'.format(device))
        else:
            # get default device, first enumerated usually
            if not devices:
                raise AudioInvalidCaptureDeviceError(
                    'No suitable audio recording devices found on this system. '
                    'Check connections and try again.')

            self._device = devices[0]  # use first

        logging.info('Using audio device #{} ({}) for audio capture'.format(
            self._device.deviceIndex, self._device.deviceName))

        # error if specified device is not suitable for capture
        if not self._device.isCapture:
            raise AudioInvalidCaptureDeviceError(
                'Specified audio device not suitable for audio recording. '
                'Has no input channels.')

        # get the sample rate
        self._sampleRateHz = \
            self._device.defaultSampleRate if sampleRateHz is None else int(
                sampleRateHz)

        logging.debug('Set stream sample rate to {} Hz'.format(
            self._sampleRateHz))

        # set the audio latency mode
        if audioLatencyMode is None:
            self._audioLatencyMode = int(prefs.hardware["audioLatencyMode"])
        else:
            self._audioLatencyMode = audioLatencyMode

        logging.debug('Set audio latency mode to {}'.format(
            self._audioLatencyMode))

        assert 0 <= self._audioLatencyMode <= 4  # sanity check for pref

        # set the number of recording channels
        self._channels = \
            self._device.inputChannels if channels is None else int(channels)

        logging.debug('Set recording channels to {} ({})'.format(
            self._channels, 'stereo' if self._channels > 1 else 'mono'))

        if self._channels > self._device.inputChannels:
            raise AudioInvalidDeviceError(
                'Invalid number of channels for audio input specified.')

        # internal recording buffer size in seconds
        assert isinstance(streamBufferSecs, (float, int))
        self._streamBufferSecs = float(streamBufferSecs)

        # PTB specific stuff
        self._mode = 2  # open a stream in capture mode

        # Handle for the recording stream, should only be opened once per
        # session
        logging.debug('Opening audio stream for device #{}'.format(
            self._device.deviceIndex))

        self._stream = audio.Stream(
            device_id=self._device.deviceIndex,
            latency_class=self._audioLatencyMode,
            mode=self._mode,
            freq=self._sampleRateHz,
            channels=self._channels)

        logging.debug('Stream opened')

        assert isinstance(audioRunMode, (float, int)) and \
               (audioRunMode == 0 or audioRunMode == 1)
        self._audioRunMode = int(audioRunMode)
        self._stream.run_mode = self._audioRunMode

        logging.debug('Set run mode to `{}`'.format(
            self._audioRunMode))

        # set latency bias
        self._stream.latency_bias = 0.0

        logging.debug('Set stream latency bias to {} ms'.format(
            self._stream.latency_bias))

        # pre-allocate recording buffer, called once
        self._stream.get_audio_data(self._streamBufferSecs)

        logging.debug(
            'Allocated stream buffer to hold {} seconds of data'.format(
                self._streamBufferSecs))

        # status flag
        self._statusFlag = NOT_STARTED

        # setup recording buffer
        self._recording = RecordingBuffer(
            sampleRateHz=self._sampleRateHz,
            channels=self._channels,
            maxRecordingSize=maxRecordingSize,
            policyWhenFull=policyWhenFull
        )

        # setup clips and transcripts dicts
        self.clips = {}
        self.lastClip = None
        self.scripts = {}
        self.lastScript = None

        logging.debug('Audio capture device #{} ready'.format(
            self._device.deviceIndex))

    @staticmethod
    def getDevices():
        """Get a `list` of audio capture device (i.e. microphones) descriptors.
        On Windows, only WASAPI devices are used.

        Returns
        -------
        list
            List of `AudioDevice` descriptors for suitable capture devices. If
            empty, no capture devices have been found.

        """
        try:
            Microphone.enforceWASAPI = bool(prefs.hardware["audioForceWASAPI"])
        except KeyError:
            pass  # use default if option not present in settings

        # query PTB for devices
        if Microphone.enforceWASAPI and sys.platform == 'win32':
            allDevs = audio.get_devices(device_type=13)
        else:
            allDevs = audio.get_devices()

        # make sure we have an array of descriptors
        allDevs = [allDevs] if isinstance(allDevs, dict) else allDevs

        # create list of descriptors only for capture devices
        inputDevices = [desc for desc in [
            AudioDeviceInfo.createFromPTBDesc(dev) for dev in allDevs]
                     if desc.isCapture]

        return inputDevices

    # def warmUp(self):
    #     """Warm-/wake-up the audio stream.
    #
    #     On some systems the first time `start` is called incurs additional
    #     latency, whereas successive calls do not. To deal with this, it is
    #     recommended that you run this warm-up routine prior to capturing audio
    #     samples. By default, this routine is called when instancing a new
    #     microphone object.
    #
    #     """
    #     # We should put an actual test here to see if timing stabilizes after
    #     # multiple invocations of this function.
    #     self._stream.start()
    #     self._stream.stop()

    @property
    def recording(self):
        """Reference to the current recording buffer (`RecordingBuffer`)."""
        return self._recording

    @property
    def recBufferSecs(self):
        """Capacity of the recording buffer in seconds (`float`)."""
        return self.recording.bufferSecs

    @property
    def maxRecordingSize(self):
        """Maximum recording size in kilobytes (`int`).

        Since audio recordings tend to consume a large amount of system memory,
        one might want to limit the size of the recording buffer to ensure that
        the application does not run out. By default, the recording buffer is
        set to 64000 KB (or 64 MB). At a sample rate of 48kHz, this will result
        in about. Using stereo audio (``nChannels == 2``) requires twice the
        buffer over mono (``nChannels == 2``) for the same length clip.

        Setting this value will allocate another recording buffer of appropriate
        size. Avoid doing this in any time sensitive parts of your application.

        """
        return self._recording.maxRecordingSize

    @maxRecordingSize.setter
    def maxRecordingSize(self, value):
        self._recording.maxRecordingSize = value

    @property
    def latencyBias(self):
        """Latency bias to add when starting the microphone (`float`).
        """
        return self._stream.latency_bias

    @latencyBias.setter
    def latencyBias(self, value):
        self._stream.latency_bias = float(value)

    @property
    def audioLatencyMode(self):
        """Audio latency mode in use (`int`). Cannot be set after
        initialization.

        """
        return self._audioLatencyMode

    @property
    def streamBufferSecs(self):
        """Size of the internal audio storage buffer in seconds (`float`).

        To ensure all data is captured, there must be less time elapsed between
        subsequent `getAudioClip` calls than `bufferSecs`.

        """
        return self._streamBufferSecs

    @property
    def status(self):
        """Status flag for the microphone. Value can be one of
        ``psychopy.constants.STARTED`` or ``psychopy.constants.NOT_STARTED``.

        For detailed stream status information, use the
        :attr:`~psychopy.sound.microphone.Microphone.streamStatus` property.

        """
        if hasattr(self, "_statusFlag"):
            return self._statusFlag

    @status.setter
    def status(self, value):
        self._statusFlag = value

    @property
    def streamStatus(self):
        """Status of the audio stream (`AudioDeviceStatus` or `None`).

        See :class:`~psychopy.sound.AudioDeviceStatus` for a complete overview
        of available status fields. This property has a value of `None` if
        the stream is presently closed.

        Examples
        --------
        Get the capture start time of the stream::

            # assumes mic.start() was called
            captureStartTime = mic.status.captureStartTime

        Check if microphone recording is active::

            isActive = mic.status.active

        Get the number of seconds recorded up to this point::

            recordedSecs = mic.status.recordedSecs

        """
        currentStatus = self._stream.status
        if currentStatus != -1:
            return AudioDeviceStatus.createFromPTBDesc(currentStatus)

    @property
    def isRecBufferFull(self):
        """`True` if there is an overflow condition with the recording buffer.

        If this is `True`, then `poll()` is still collecting stream samples but
        is no longer writing them to anything, causing stream samples to be
        lost.

        """
        return self._recording.isFull

    @property
    def isStarted(self):
        """``True`` if stream recording has been started (`bool`)."""
        return self.status == STARTED

    def start(self, when=None, waitForStart=0, stopTime=None):
        """Start an audio recording.

        Calling this method will begin capturing samples from the microphone and
        writing them to the buffer.

        Parameters
        ----------
        when : float, int or None
            When to start the stream. If the time specified is a floating point
            (absolute) system time, the device will attempt to begin recording
            at that time. If `None` or zero, the system will try to start
            recording as soon as possible.
        waitForStart : bool
            Wait for sound onset if `True`.
        stopTime : float, int or None
            Number of seconds to record. If `None` or `-1`, recording will
            continue forever until `stop` is called.

        Returns
        -------
        float
            Absolute time the stream was started.

        """
        # check if the stream has been
        if self.isStarted:
            raise AudioStreamError(
                "Cannot start a stream, already started.")

        if self._stream is None:
            raise AudioStreamError("Stream not ready.")

        # reset the writing 'head'
        self._recording.seek(0, absolute=True)

        # reset warnings
        # self._warnedRecBufferFull = False

        startTime = self._stream.start(
            repetitions=0,
            when=when,
            wait_for_start=int(waitForStart),
            stop_time=stopTime)

        # recording has begun or is scheduled to do so
        self._statusFlag = STARTED

        logging.debug(
            'Scheduled start of audio capture for device #{} at t={}.'.format(
                self._device.deviceIndex, startTime))

        return startTime

    def record(self, when=None, waitForStart=0, stopTime=None):
        """Start an audio recording (alias of `.start()`).

        Calling this method will begin capturing samples from the microphone and
        writing them to the buffer.

        Parameters
        ----------
        when : float, int or None
            When to start the stream. If the time specified is a floating point
            (absolute) system time, the device will attempt to begin recording
            at that time. If `None` or zero, the system will try to start
            recording as soon as possible.
        waitForStart : bool
            Wait for sound onset if `True`.
        stopTime : float, int or None
            Number of seconds to record. If `None` or `-1`, recording will
            continue forever until `stop` is called.

        Returns
        -------
        float
            Absolute time the stream was started.

        """
        return self.start(
            when=when,
            waitForStart=waitForStart,
            stopTime=stopTime)

    def stop(self, blockUntilStopped=True, stopTime=None):
        """Stop recording audio.

        Call this method to end an audio recording if in progress. This will
        simply halt recording and not close the stream. Any remaining samples
        will be polled automatically and added to the recording buffer.

        Parameters
        ----------
        blockUntilStopped : bool
            Halt script execution until the stream has fully stopped.
        stopTime : float or None
            Scheduled stop time for the stream in system time. If `None`, the
            stream will stop as soon as possible.

        Returns
        -------
        tuple
            Tuple containing `startTime`, `endPositionSecs`, `xruns` and
            `estStopTime`.

        """
        if not self.isStarted:
            raise AudioStreamError(
                "Cannot stop a stream that has not been started.")

        # poll remaining samples, if any
        if not self.isRecBufferFull:
            self.poll()

        startTime, endPositionSecs, xruns, estStopTime = self._stream.stop(
            block_until_stopped=int(blockUntilStopped),
            stopTime=stopTime)
        self._statusFlag = NOT_STARTED

        logging.debug(
            ('Device #{} stopped capturing audio samples at estimated time '
             't={}. Total overruns: {} Total recording time: {}').format(
                self._device.deviceIndex, estStopTime, xruns, endPositionSecs))

        return startTime, endPositionSecs, xruns, estStopTime

    def pause(self, blockUntilStopped=True, stopTime=None):
        """Pause a recording (alias of `.stop`).

        Call this method to end an audio recording if in progress. This will
        simply halt recording and not close the stream. Any remaining samples
        will be polled automatically and added to the recording buffer.

        Parameters
        ----------
        blockUntilStopped : bool
            Halt script execution until the stream has fully stopped.
        stopTime : float or None
            Scheduled stop time for the stream in system time. If `None`, the
            stream will stop as soon as possible.

        Returns
        -------
        tuple
            Tuple containing `startTime`, `endPositionSecs`, `xruns` and
            `estStopTime`.

        """
        return self.stop(blockUntilStopped=blockUntilStopped, stopTime=stopTime)

    def close(self):
        """Close the stream.

        Should not be called until you are certain you're done with it. Ideally,
        you should never close and reopen the same stream within a single
        session.

        """
        self._stream.close()
        logging.debug('Stream closed')

    def poll(self):
        """Poll audio samples.

        Calling this method adds audio samples collected from the stream buffer
        to the recording buffer that have been captured since the last `poll`
        call. Time between calls of this function should be less than
        `bufferSecs`. You do not need to call this if you call `stop` before
        the time specified by `bufferSecs` elapses since the `start` call.

        Can only be called between called of `start` (or `record`) and `stop`
        (or `pause`).

        Returns
        -------
        int
            Number of overruns in sampling.

        """
        if not self.isStarted:
            raise AudioStreamError(
                "Cannot poll samples from audio device, not started.")

        # figure out what to do with this other information
        audioData, absRecPosition, overflow, cStartTime = \
            self._stream.get_audio_data()

        if overflow:
            logging.warning(
                "Audio stream buffer overflow, some audio samples have been "
                "lost! To prevent this, ensure `Microphone.poll()` is being "
                "called often enough, or increase the size of the audio buffer "
                "with `bufferSecs`.")

        overruns = self._recording.write(audioData)

        return overruns

    def bank(self, tag=None, transcribe=False, **kwargs):
        """Store current buffer as a clip within the microphone object.

        This method is used internally by the Microphone component in Builder,
        don't use it for other applications. Either `stop()` or `pause()` must
        be called before calling this method.

        Parameters
        ----------
        tag : str or None
            Label for the clip.
        transcribe : bool or str
            Set to the name of a transcription engine (e.g. "GOOGLE") to
            transcribe using that engine, or set as `False` to not transcribe.
        kwargs : dict
            Additional keyword arguments to pass to
            :class:`~psychopy.sound.AudioClip.transcribe()`.

        """
        # make sure the tag exists in both clips and transcripts dicts
        if tag not in self.clips:
            self.clips[tag] = []
        if tag not in self.scripts:
            self.scripts[tag] = []
        # append current recording to clip list according to tag
        self.lastClip = self.getRecording()
        self.clips[tag].append(self.lastClip)
        # synonymise null values
        if transcribe in ('undefined', 'NONE', 'None', 'none', 'False', 'false', 'FALSE'):
            transcribe = False
        # append current clip's transcription according to tag
        if transcribe:
            if transcribe in ('Built-in', True, 'BUILT_IN', 'BUILT-IN',
                              'Built-In', 'built-in'):
                engine = "sphinx"
            elif type(transcribe) == str:
                engine = transcribe
            self.lastScript = self.lastClip.transcribe(
                engine=engine, **kwargs)
        else:
            self.lastScript = "Transcription disabled."
        self.scripts[tag].append(self.lastScript)
        # clear recording buffer
        self._recording.clear()
        # return banked items
        if transcribe:
            return self.lastClip, self.lastScript
        else:
            return self.lastClip

    def clear(self):
        """Wipe all clips. Deletes previously banked audio clips.
        """
        # clear clips
        self.clips = {}
        # clear recording
        self._recording.clear()

    def flush(self):
        """Get a copy of all banked clips, then clear the clips from storage."""
        # get copy of clips dict
        clips = self.clips.copy()
        # clear
        self.clear()

        return clips

    def getRecording(self):
        """Get audio data from the last microphone recording.

        Call this after `stop` to get the recording as an `AudioClip` object.
        Raises an error if a recording is in progress.

        Returns
        -------
        AudioClip
            Recorded data between the last calls to `start` (or `record`) and
            `stop`.

        """
        if self.isStarted:
            raise AudioStreamError(
                "Cannot get audio clip, recording was in progress. Be sure to "
                "call `Microphone.stop` first.")

        return self._recording.getSegment()  # full recording


if __name__ == "__main__":
    pass
